from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import albumentations as albu
import numpy as np
import pandas as pd
import torchvision
from torch import FloatTensor

from oml.const import (
    BLACK,
    INDEX_KEY,
    INPUT_TENSORS_KEY,
    LABELS_COLUMN,
    LABELS_KEY,
    PATHS_COLUMN,
    SPLIT_COLUMN,
    X1_COLUMN,
    X2_COLUMN,
    Y1_COLUMN,
    Y2_COLUMN,
    TBBoxes,
    TColor,
)
from oml.datasets.dataframe import (
    DFLabeledDataset,
    DFQueryGalleryDataset,
    DFQueryGalleryLabeledDataset,
)
from oml.interfaces.datasets import (
    IBaseDataset,
    ILabeledDataset,
    IQueryGalleryLabeledDataset,
    IVisualizableDataset,
)
from oml.registry.transforms import get_transforms
from oml.transforms.images.utils import TTransforms, get_im_reader_for_transforms
from oml.utils.dataframe_format import check_retrieval_dataframe_format
from oml.utils.images.images import TImReader, draw_bbox, imread_pillow


def parse_bboxes(df: pd.DataFrame) -> Optional[TBBoxes]:
    n_existing_columns = sum([x in df for x in [X1_COLUMN, Y1_COLUMN, X2_COLUMN, Y2_COLUMN]])

    if n_existing_columns == 4:
        bboxes = []
        for _, row in df.iterrows():
            bbox = int(row[X1_COLUMN]), int(row[Y1_COLUMN]), int(row[X2_COLUMN]), int(row[Y2_COLUMN])
            bbox = None if any(coord is None for coord in bbox) else bbox
            bboxes.append(bbox)

    elif n_existing_columns == 0:
        bboxes = None

    else:
        raise ValueError(f"Found {n_existing_columns} bounding bboxes columns instead of 4. Check your dataframe.")

    return bboxes


class ImageBaseDataset(IBaseDataset, IVisualizableDataset):
    """
    The base class that handles image specific logic.

    """

    def __init__(
        self,
        paths: List[Path],
        dataset_root: Optional[Union[str, Path]] = None,
        bboxes: Optional[TBBoxes] = None,
        extra_data: Optional[Dict[str, Any]] = None,
        transform: Optional[TTransforms] = None,
        f_imread: Optional[TImReader] = None,
        cache_size: Optional[int] = 0,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        index_key: str = INDEX_KEY,
    ):
        """

        Args:
            paths: Paths to images. Will be concatenated with ``dataset_root`` if provided.
            dataset_root: Path to the images' dir, set ``None`` if you provided the absolute paths in your dataframe
            bboxes: Bounding boxes of images. Some of the images may not have bounding bboxes.
            extra_data: Dictionary containing records of some additional information.
            transform: Augmentations for the images, set ``None`` to perform only normalisation and casting to tensor
            f_imread: Function to read the images, pass ``None`` to pick it automatically based on provided transforms
            cache_size: Size of the dataset's cache
            input_tensors_key: Key to put tensors into the batches
            index_key: Key to put samples' ids into the batches

        """
        assert (bboxes is None) or (len(paths) == len(bboxes))

        if extra_data is not None:
            assert all(
                len(record) == len(paths) for record in extra_data.values()
            ), "All the extra records need to have the size equal to the dataset's size"
            self.extra_data = extra_data
        else:
            self.extra_data = {}

        self.input_tensors_key = input_tensors_key
        self.index_key = index_key

        if dataset_root is not None:
            paths = list(map(lambda x: Path(dataset_root) / x, paths))

        self._paths = list(map(str, paths))
        self._bboxes = bboxes
        self._transform = transform if transform else get_transforms("norm_albu")
        self._f_imread = f_imread or get_im_reader_for_transforms(self._transform)

        if cache_size:
            self.read_bytes = lru_cache(maxsize=cache_size)(self._read_bytes)  # type: ignore
        else:
            self.read_bytes = self._read_bytes  # type: ignore

        available_transforms = (albu.Compose, torchvision.transforms.Compose)
        assert isinstance(self._transform, available_transforms), f"Transforms must one of: {available_transforms}"

    @staticmethod
    def _read_bytes(path: Union[Path, str]) -> bytes:
        with open(str(path), "rb") as fin:
            return fin.read()

    def __getitem__(self, item: int) -> Dict[str, Union[FloatTensor, int]]:
        img_bytes = self.read_bytes(self._paths[item])
        img = self._f_imread(img_bytes)

        im_h, im_w = img.shape[:2] if isinstance(img, np.ndarray) else img.size[::-1]

        if (self._bboxes is not None) and (self._bboxes[item] is not None):
            x1, y1, x2, y2 = self._bboxes[item]
        else:
            x1, y1, x2, y2 = 0, 0, im_w, im_h

        if isinstance(self._transform, albu.Compose):
            img = img[y1:y2, x1:x2, :]
            image_tensor = self._transform(image=img)["image"]
        else:
            # torchvision.transforms
            img = img.crop((x1, y1, x2, y2))
            image_tensor = self._transform(img)

        data = {
            self.input_tensors_key: image_tensor,
            self.index_key: item,
        }

        for key, record in self.extra_data.items():
            if key in data:
                raise ValueError(f"<extra_data> and dataset share the same key: {key}")
            else:
                data[key] = record[item]

        return data

    def __len__(self) -> int:
        return len(self._paths)

    def visualize(self, item: int, color: TColor = BLACK) -> np.ndarray:
        img = np.array(imread_pillow(self.read_bytes(self._paths[item])))
        bbox = self._bboxes[item] if (self._bboxes is not None) else None
        image = draw_bbox(im=img, bbox=bbox, color=color)

        return image


class ImageLabeledDataset(DFLabeledDataset, IVisualizableDataset):
    """
    The dataset of images having their ground truth labels.

    """

    _dataset: ImageBaseDataset

    def __init__(
        self,
        df: pd.DataFrame,
        extra_data: Optional[Dict[str, Any]] = None,
        dataset_root: Optional[Union[str, Path]] = None,
        transform: Optional[albu.Compose] = None,
        f_imread: Optional[TImReader] = None,
        cache_size: Optional[int] = 0,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        labels_key: str = LABELS_KEY,
        index_key: str = INDEX_KEY,
    ):
        dataset = ImageBaseDataset(
            paths=df[PATHS_COLUMN].tolist(),
            bboxes=parse_bboxes(df),
            extra_data=None,
            dataset_root=dataset_root,
            transform=transform,
            f_imread=f_imread,
            cache_size=cache_size,
            input_tensors_key=input_tensors_key,
            index_key=index_key,
        )
        super().__init__(dataset=dataset, df=df, extra_data=extra_data, labels_key=labels_key)

    def visualize(self, item: int, color: TColor) -> np.ndarray:
        return self._dataset.visualize(item=item, color=color)


class ImageQueryGalleryLabeledDataset(DFQueryGalleryLabeledDataset, IVisualizableDataset):
    """
    The annotated dataset of images having `query`/`gallery` split.

    Note, that some datasets used as benchmarks in Metric Learning
    explicitly provide the splitting information (for example, ``DeepFashion InShop`` dataset), but some of them
    don't (for example, ``CARS196`` or ``CUB200``). The validation idea for the latter is to perform `1 vs rest`
    validation, where every query is evaluated versus the whole validation dataset (except for this exact query).

    So, if you want an item participate in validation as both: query and gallery, you should mark this item as
    ``is_query == True`` and ``is_gallery == True``, as it's done in the `CARS196` or `CUB200` dataset.

    """

    _dataset: ImageBaseDataset

    def __init__(
        self,
        df: pd.DataFrame,
        extra_data: Optional[Dict[str, Any]] = None,
        dataset_root: Optional[Union[str, Path]] = None,
        transform: Optional[albu.Compose] = None,
        f_imread: Optional[TImReader] = None,
        cache_size: Optional[int] = 0,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        labels_key: str = LABELS_KEY,
    ):
        dataset = ImageBaseDataset(
            paths=df[PATHS_COLUMN].tolist(),
            bboxes=parse_bboxes(df),
            extra_data=None,
            dataset_root=dataset_root,
            transform=transform,
            f_imread=f_imread,
            cache_size=cache_size,
            input_tensors_key=input_tensors_key,
        )

        super().__init__(dataset=dataset, df=df, extra_data=extra_data, labels_key=labels_key)

    def visualize(self, item: int, color: TColor) -> np.ndarray:
        return self._dataset.visualize(item=item, color=color)


class ImageQueryGalleryDataset(DFQueryGalleryDataset, IVisualizableDataset):
    """
    The NOT annotated dataset of images having `query`/`gallery` split.

    """

    _dataset: ImageBaseDataset

    def __init__(
        self,
        df: pd.DataFrame,
        extra_data: Optional[Dict[str, Any]] = None,
        dataset_root: Optional[Union[str, Path]] = None,
        transform: Optional[albu.Compose] = None,
        f_imread: Optional[TImReader] = None,
        cache_size: Optional[int] = 0,
        input_tensors_key: str = INPUT_TENSORS_KEY,
    ):
        dataset = ImageBaseDataset(
            paths=df[PATHS_COLUMN].tolist(),
            bboxes=parse_bboxes(df),
            extra_data=None,
            dataset_root=dataset_root,
            transform=transform,
            f_imread=f_imread,
            cache_size=cache_size,
            input_tensors_key=input_tensors_key,
        )
        super().__init__(dataset=dataset, df=df, extra_data=extra_data)

    def visualize(self, item: int, color: TColor) -> np.ndarray:
        return self._dataset.visualize(item=item, color=color)


def get_retrieval_images_datasets(
    dataset_root: Path,
    transforms_train: Any,
    transforms_val: Any,
    f_imread_train: Optional[TImReader] = None,
    f_imread_val: Optional[TImReader] = None,
    dataframe_name: str = "df.csv",
    cache_size: Optional[int] = 0,
    verbose: bool = True,
) -> Tuple[ILabeledDataset, IQueryGalleryLabeledDataset]:
    df = pd.read_csv(dataset_root / dataframe_name, index_col=False)

    check_retrieval_dataframe_format(df, dataset_root=dataset_root, verbose=verbose)

    # first half will consist of "train" split, second one of "val"
    # so labels in train will be from 0 to N-1 and labels in test will be from N to K
    mapper = {l: i for i, l in enumerate(df.sort_values(by=[SPLIT_COLUMN])[LABELS_COLUMN].unique())}

    # train
    df_train = df[df[SPLIT_COLUMN] == "train"].reset_index(drop=True)
    df_train[LABELS_COLUMN] = df_train[LABELS_COLUMN].map(mapper)

    train_dataset = ImageLabeledDataset(
        df=df_train,
        dataset_root=dataset_root,
        transform=transforms_train,
        cache_size=cache_size,
        f_imread=f_imread_train,
    )

    # val (query + gallery)
    df_query_gallery = df[df[SPLIT_COLUMN] == "validation"].reset_index(drop=True)
    valid_dataset = ImageQueryGalleryLabeledDataset(
        df=df_query_gallery,
        dataset_root=dataset_root,
        transform=transforms_val,
        cache_size=cache_size,
        f_imread=f_imread_val,
    )

    return train_dataset, valid_dataset


__all__ = [
    "ImageBaseDataset",
    "ImageLabeledDataset",
    "ImageQueryGalleryDataset",
    "ImageQueryGalleryLabeledDataset",
    "get_retrieval_images_datasets",
]



from typing import Dict, Optional, Sequence, Union

import pytorch_lightning as pl
from torch.utils.data import DataLoader

from oml.ddp.patching import patch_dataloader_to_ddp

TValDataloaders = Union[DataLoader, Sequence[DataLoader]]
TTrainDataloaders = Union[TValDataloaders, Dict[str, DataLoader]]


class ModuleDDP(pl.LightningModule):
    """
    The module automatically patches training and validation dataloaders to DDP mode by splitting available indices
    between devices. Note, don't use ``trainer.fit(...)`` or ``trainer.validate(...)``, because in this case,
    `PytorchLightning` will ignore our patching.

    """

    def __init__(
        self,
        loaders_train: Optional[TTrainDataloaders] = None,
        loaders_val: Optional[TValDataloaders] = None,
    ):
        assert loaders_train is not None or loaders_val is not None, "At least one dataloader must be specified"
        pl.LightningModule.__init__(self)
        self.loaders_train = loaders_train
        self.loaders_val = loaders_val

    def train_dataloader(self) -> TTrainDataloaders:
        return self._patch_loaders("train") if self.loaders_train else super(ModuleDDP, self).train_dataloader()

    def val_dataloader(self) -> TValDataloaders:
        return self._patch_loaders("val") if self.loaders_val else super(ModuleDDP, self).val_dataloader()

    def _patch_loaders(self, mode: str) -> TTrainDataloaders:
        assert mode in ("train", "val")
        loaders = self.loaders_train if mode == "train" else self.loaders_val

        if isinstance(loaders, DataLoader):
            return patch_dataloader_to_ddp(loaders)
        elif isinstance(loaders, (list, tuple)):
            return [patch_dataloader_to_ddp(loader) for loader in loaders]
        elif isinstance(loaders, dict):
            return {k: patch_dataloader_to_ddp(loader) for k, loader in loaders.items()}
        else:
            raise TypeError("Not supported loaders type")


__all__ = ["ModuleDDP"]



from oml.datasets.images import ImageLabeledDataset, ImageQueryGalleryLabeledDataset


class DatasetWithLabels(ImageLabeledDataset):
    # this class allows to have backward compatibility
    pass


class DatasetQueryGallery(ImageQueryGalleryLabeledDataset):
    # this class allows to have backward compatibility
    pass


__all__ = ["DatasetWithLabels", "DatasetQueryGallery"]



import warnings
from copy import deepcopy
from pprint import pformat
from typing import Callable, List, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from torch import FloatTensor, LongTensor

from oml.const import (
    BLACK,
    BLUE,
    GRAY,
    GREEN,
    N_GT_SHOW_EMBEDDING_METRICS,
    RED,
    SEQUENCE_COLUMN,
    TColor,
)
from oml.functional.knn import batched_knn, batched_knn_qg
from oml.interfaces.datasets import (
    IBaseDataset,
    ILabeledDataset,
    IQueryGalleryDataset,
    IQueryGalleryLabeledDataset,
    IVisualizableDataset,
)
from oml.utils.misc_torch import is_sorted_tensor


def get_sequence_from_dataset(dataset: IBaseDataset) -> Optional[LongTensor]:
    if SEQUENCE_COLUMN in dataset.extra_data:
        sequence = pd.Series(dataset.extra_data[SEQUENCE_COLUMN])
        sequence_ids = LongTensor(pd.factorize(sequence, sort=True)[0])
    else:
        sequence_ids = None

    return sequence_ids


class RetrievalResults:
    _max_elements_in_str_repr: int = 100

    def __init__(
        self,
        distances: Sequence[FloatTensor],
        retrieved_ids: Sequence[LongTensor],
        gt_ids: Optional[Sequence[LongTensor]] = None,
    ):
        """
        Args:
            distances: Sorted distances from queries to the first gallery items with the size of ``n_query``.
            retrieved_ids: First gallery indices retrieved for every query with the size of ``n_query``.
                Every index is within the range ``(0, n_gallery - 1)``.
            gt_ids: Gallery indices relevant to every query with the size of ``n_query``.
                Every element is within the range ``(0, n_gallery - 1)``

        """
        for d, r in zip(distances, retrieved_ids):
            if not is_sorted_tensor(d):
                raise RuntimeError(f"Distances must be sorted: {d}.")
            if not len(torch.unique(r[:100])) == len(r[:100]):  # it's too expensive to check all the ids!
                raise RuntimeError("Retrieved ids must be unique.")
            if not len(d) == len(r):
                raise RuntimeError("Number of distances and retrieved items must match.")
            if (d.ndim != 1) or (r.ndim != 1):
                raise RuntimeError("Distances and retrieved items must be 1-d tensors.")

        if gt_ids is not None:
            assert len(distances) == len(gt_ids)
            if any(len(x) == 0 for x in gt_ids):
                warnings.warn("Some of the queries don't have available gts.")

        self._distances = tuple(distances)
        self._retrieved_ids = tuple(retrieved_ids)
        self._gt_ids = tuple(gt_ids) if gt_ids is not None else None

    @property
    def distances(self) -> Tuple[FloatTensor, ...]:
        """
        Returns:
            Sorted distances from queries to the first gallery items with the size of ``n_query``.
        """
        return self._distances

    @property
    def retrieved_ids(self) -> Tuple[LongTensor, ...]:
        """
        Returns:
            First gallery indices retrieved for every query with the size of ``n_query``.
            Every index is within the range ``(0, n_gallery - 1)``.
        """
        return self._retrieved_ids

    @property
    def gt_ids(self) -> Optional[Tuple[LongTensor, ...]]:
        """
        Returns:
            Gallery indices relevant to every query with the size of ``n_query``.
            Every element is within the range ``(0, n_gallery - 1)``
        """
        return self._gt_ids

    @property
    def n_retrieved_items(self) -> int:
        """
        Returns:
            Number of items retrieved for each query. If queries have different number of retrieved items,
            returns the maximum of them.

        """
        return max(len(x) for x in self.retrieved_ids)

    def is_empty(self) -> bool:
        return all(len(rids) == 0 for rids in self.retrieved_ids)

    def deepcopy(self) -> "RetrievalResults":
        return RetrievalResults(
            retrieved_ids=deepcopy(self.retrieved_ids), distances=deepcopy(self.distances), gt_ids=deepcopy(self.gt_ids)
        )

    @classmethod
    def from_embeddings(
        cls,
        embeddings: FloatTensor,
        dataset: Union[IQueryGalleryDataset, IQueryGalleryLabeledDataset],
        n_items: int = 100,
        verbose: bool = False,
    ) -> "RetrievalResults":
        """
        Args:
            embeddings: The result of inference with the shape of ``[dataset_len, emb_dim]``.
            dataset: Dataset having query/gallery split.
            n_items: Number of the closest gallery items to retrieve. It may be clipped by
                gallery size if needed. Note, some queries may get less than this number of retrieved items if they
                don't have enough gallery items available.
            verbose: Set ``True`` to see progress bar.

        """
        assert len(embeddings) == len(dataset), "Embeddings and dataset must have the same size."

        sequence_ids = get_sequence_from_dataset(dataset)

        labels_gt = dataset.get_labels() if isinstance(dataset, IQueryGalleryLabeledDataset) else None

        distances, retrieved_ids, gt_ids = batched_knn(
            embeddings=embeddings,
            ids_query=dataset.get_query_ids(),
            ids_gallery=dataset.get_gallery_ids(),
            labels_gt=labels_gt,
            sequence_ids=sequence_ids,
            top_n=n_items,
            verbose=verbose,
        )

        return RetrievalResults(distances=distances, retrieved_ids=retrieved_ids, gt_ids=gt_ids)

    @classmethod
    def from_embeddings_qg(
        cls,
        embeddings_query: FloatTensor,
        embeddings_gallery: FloatTensor,
        dataset_query: Union[IBaseDataset, ILabeledDataset],
        dataset_gallery: Union[IBaseDataset, ILabeledDataset],
        n_items: int = 100,
        verbose: bool = False,
    ) -> "RetrievalResults":
        """
        Args:
            embeddings_query: The result of inference with the shape of ``[n_queries, emb_dim]``.
            embeddings_gallery: The result of inference with the shape of ``[n_galleries, emb_dim]``.
            dataset_query: Dataset of queries with the length of ``n_queries``.
            dataset_gallery: Dataset of galleries with the length of ``n_galleries``.
            n_items: Number of the closest gallery items to retrieve. It may be clipped by
                gallery size if needed. Note, some queries may get less than this number of retrieved items if they
                don't have enough gallery items available.
            verbose: Set ``True`` to see progress bar.

        """
        assert len(embeddings_query) == len(dataset_query), "Embeddings and dataset must have the same size."
        assert len(embeddings_gallery) == len(dataset_gallery), "Embeddings and dataset must have the same size."

        labels_query = dataset_query.get_labels() if isinstance(dataset_query, ILabeledDataset) else None
        labels_gallery = dataset_gallery.get_labels() if isinstance(dataset_gallery, ILabeledDataset) else None

        sequence_ids_query = get_sequence_from_dataset(dataset_query)
        sequence_ids_gallery = get_sequence_from_dataset(dataset_gallery)

        distances, retrieved_ids, gt_ids = batched_knn_qg(
            embeddings_query=embeddings_query,
            embeddings_gallery=embeddings_gallery,
            ids_query=None,
            ids_gallery=None,
            labels_query=labels_query,
            labels_gallery=labels_gallery,
            sequence_ids_query=sequence_ids_query,
            sequence_ids_gallery=sequence_ids_gallery,
            top_n=n_items,
            verbose=verbose,
        )

        return RetrievalResults(distances=distances, retrieved_ids=retrieved_ids, gt_ids=gt_ids)

    def __str__(self) -> str:
        m = self._max_elements_in_str_repr

        txt = (
            f"Maximum number of retrieved items: {self.n_retrieved_items}.\n"
            f"Distances to the retrieved items:\n{pformat(self.distances[:m])}.\n"
            f"Ids of the retrieved gallery items:\n{pformat(self.retrieved_ids[:m])}.\n"
        )

        if self.gt_ids is None:
            txt += "Ground truths are unknown.\n"
        else:
            gt_ids_list = [x.tolist() for x in self.gt_ids]
            txt += f"Ground truth gallery ids are:\n{pformat(gt_ids_list[:m])}.\n"

        return txt

    def visualize_qg(
        self,
        query_ids: List[int],
        dataset_query: IVisualizableDataset,
        dataset_gallery: IVisualizableDataset,
        n_galleries_to_show: int = 5,
        n_gt_to_show: int = N_GT_SHOW_EMBEDDING_METRICS,
        verbose: bool = False,
        show: bool = False,
    ) -> plt.Figure:
        """
        Args:
            query_ids: Query indices within the range of ``(0, n_query - 1)``.
            dataset_query: Dataset of queries supporting visualisation, with the length of ``n_query``.
            dataset_gallery: Dataset of queries supporting visualisation, with the length of ``n_gallery``.
            n_galleries_to_show: Number of closest gallery items to show.
            n_gt_to_show: Number of ground truth gallery items to show for reference (if available).
            verbose: Set ``True`` to allow prints.
            show: Set ``True`` to instantly visualise the resulted figure.

        """
        dq_name = dataset_query.__class__.__name__
        dg_name = dataset_gallery.__class__.__name__

        if not isinstance(dataset_query, IVisualizableDataset):
            raise TypeError(f"Query dataset has to support {IVisualizableDataset.__name__}. Got {dq_name}.")

        if not isinstance(dataset_gallery, IVisualizableDataset):
            raise TypeError(f"Gallery dataset has to support {IVisualizableDataset.__name__}. Got {dg_name}.")

        nq1, nq2 = len(self.retrieved_ids), len(dataset_query)
        if nq1 != nq2:
            raise RuntimeError(
                f"Number of queries in {self.__class__.__name__} and {dq_name} must match: {nq1} != {nq2}"
            )

        if verbose:
            print(f"Visualizing {n_galleries_to_show} for the following query ids: {query_ids}.")

        def visualize_query_fn(item: int, color: TColor) -> np.ndarray:
            return dataset_query.visualize(item=item, color=color)

        def visualize_gallery_fn(item: int, color: TColor) -> np.ndarray:
            return dataset_gallery.visualize(item=item, color=color)

        return self.visualize_with_functions(
            query_ids=query_ids,
            visualize_query_fn=visualize_query_fn,
            visualize_gallery_fn=visualize_gallery_fn,
            n_galleries_to_show=n_galleries_to_show,
            n_gt_to_show=n_gt_to_show,
            show=show,
        )

    def visualize(
        self,
        query_ids: List[int],
        dataset: IQueryGalleryDataset,
        n_galleries_to_show: int = 5,
        n_gt_to_show: int = N_GT_SHOW_EMBEDDING_METRICS,
        verbose: bool = False,
        show: bool = False,
    ) -> plt.Figure:
        """
        Args:
            query_ids: Query indices within the range of ``(0, n_query - 1)``.
            dataset: Dataset that provides query-gallery split and supports visualisation.
            n_galleries_to_show: Number of closest gallery items to show.
            n_gt_to_show: Number of ground truth gallery items to show for reference (if available).
            verbose: Set ``True`` to allow prints.
            show: Set ``True`` to instantly visualise the resulted figure.

        """
        dataset_name = dataset.__class__.__name__
        if not isinstance(dataset, IVisualizableDataset):
            raise TypeError(f"Dataset has to support {IVisualizableDataset.__name__}. Got {dataset_name}.")
        if not isinstance(dataset, IQueryGalleryDataset):
            raise TypeError(f"Dataset has to support {IQueryGalleryDataset.__name__}. Got {dataset_name}.")

        nq1, nq2 = len(self.retrieved_ids), len(dataset.get_query_ids())
        if nq1 != nq2:
            raise RuntimeError(
                f"Number of queries in {self.__class__.__name__} and {dataset_name} " f"must match: {nq1} != {nq2}"
            )

        if verbose:
            print(f"Visualizing {n_galleries_to_show} for the following query ids: {query_ids}.")

        def visualize_query_fn(item: int, color: TColor) -> np.ndarray:
            return dataset.visualize(item=dataset.get_query_ids()[item].item(), color=color)

        def visualize_gallery_fn(item: int, color: TColor) -> np.ndarray:
            return dataset.visualize(item=dataset.get_gallery_ids()[item].item(), color=color)

        return self.visualize_with_functions(
            query_ids=query_ids,
            visualize_query_fn=visualize_query_fn,
            visualize_gallery_fn=visualize_gallery_fn,
            n_galleries_to_show=n_galleries_to_show,
            n_gt_to_show=n_gt_to_show,
            show=show,
        )

    def visualize_with_functions(
        self,
        query_ids: List[int],
        visualize_query_fn: Callable[[int, TColor], np.ndarray],
        visualize_gallery_fn: Callable[[int, TColor], np.ndarray],
        n_galleries_to_show: int = 5,
        n_gt_to_show: int = N_GT_SHOW_EMBEDDING_METRICS,
        show: bool = False,
    ) -> plt.Figure:
        """
        Args:
            query_ids: Query indices within the range of ``(0, n_query - 1)``.
            visualize_query_fn: Function plotting ``i-th`` query with respect to the given color.
            visualize_gallery_fn: Function plotting ``j-th`` gallery with respect to the given color.
            n_galleries_to_show: Number of closest gallery items to show.
            n_gt_to_show: Number of ground truth gallery items to show for reference (if available).
            show: Set ``True`` to instantly visualize the resulted figure.

        """

        max_presented_galleries = max(len(self.retrieved_ids[iq]) for iq in query_ids)
        n_galleries_to_show = min(n_galleries_to_show, max_presented_galleries)
        n_gt_to_show = n_gt_to_show if (self.gt_ids is not None) else 0

        fig = plt.figure(figsize=(16, 16 / (n_galleries_to_show + n_gt_to_show + 1) * len(query_ids)))
        n_rows, n_cols = len(query_ids), n_galleries_to_show + 1 + n_gt_to_show

        # iterate over queries
        for i, query_idx in enumerate(query_ids):

            plt.subplot(n_rows, n_cols, i * (n_galleries_to_show + 1 + n_gt_to_show) + 1)

            img = visualize_query_fn(query_idx, BLUE)

            plt.imshow(img)
            plt.title(f"Query #{query_idx}")
            plt.axis("off")

            # iterate over retrieved items
            for j, ret_idx in enumerate(self.retrieved_ids[query_idx][:n_galleries_to_show]):
                if self.gt_ids is not None:
                    color = GREEN if ret_idx in self.gt_ids[query_idx] else RED
                else:
                    color = BLACK

                plt.subplot(n_rows, n_cols, i * (n_galleries_to_show + 1 + n_gt_to_show) + j + 2)
                img = visualize_gallery_fn(ret_idx, color)

                plt.title(f"Gallery #{ret_idx} - {round(self.distances[query_idx][j].item(), 3)}")
                plt.imshow(img)
                plt.axis("off")

            if self.gt_ids is not None:

                for k, gt_idx in enumerate(self.gt_ids[query_idx][:n_gt_to_show]):
                    plt.subplot(
                        n_rows, n_cols, i * (n_galleries_to_show + 1 + n_gt_to_show) + k + n_galleries_to_show + 2
                    )

                    img = visualize_gallery_fn(gt_idx, GRAY)

                    plt.title("GT")
                    plt.imshow(img)
                    plt.axis("off")

        fig.tight_layout()

        if show:
            fig.show()

        return fig


__all__ = ["RetrievalResults"]



from itertools import combinations, product
from operator import itemgetter
from random import sample
from sys import maxsize
from typing import List

import numpy as np
import torch

from oml.interfaces.miners import ITripletsMinerInBatch, TLabels, TTripletsIds
from oml.utils.misc import find_value_ids


class AllTripletsMiner(ITripletsMinerInBatch):
    """
    This miner selects all the possible triplets for the given batch.

    """

    def __init__(self, max_output_triplets: int = maxsize, device: str = "cpu"):
        """
        Args:
            max_output_triplets: Number of all of the possible triplets
              in the batch can be very large, so we can limit them vis this parameter.

            device: the device where to perform computations.

        """
        self._max_out_triplets = max_output_triplets
        self._device = device

    def _sample(self, *_: torch.Tensor, labels: List[int]) -> TTripletsIds:  # type: ignore
        """
        Args:
            labels: Labels of the samples in the batch
            *_: Note, that we ignore features argument

        Returns:
            Indices of the triplets

        """
        return get_available_triplets(labels, max_out_triplets=self._max_out_triplets, device=self._device)


def get_available_triplets(labels: TLabels, max_out_triplets: int = maxsize, device: str = "cpu") -> TTripletsIds:
    """
    Generate a random subset of size ``max_out_triplets`` from the set of
    all possible triplets (anchor, positive, negative).

    Args:
        labels: Labels of the samples.
        max_out_triplets: maximal number of triplets to generate
        device: the desired device of returned tensor

    Returns:
        lists of indexes of the anchors, positives and negatives.

    """
    labels = torch.tensor(labels, device=device, requires_grad=False)

    # First we need to get initial ingridients of triplets:
    # the arrays of all (anchor, positive) and (anchor, negavite) pairs
    mask_same_label: torch.Tensor = labels[:, None] == labels[None, :]
    idx_anch_neg = torch.nonzero(torch.logical_not(mask_same_label))
    mask_same_label.fill_diagonal_(False)
    idx_anch_pos = torch.nonzero(mask_same_label)
    del mask_same_label

    # Next we need to group (anchor, negative) pairs
    # with respect to the anchor. As a result we get an array called neg_groups,
    # such that neg_groups[i] is all the negatives, that come along with
    # the anchor number i.
    # For simplicity, lets assume that
    # idx_anch_pos = [[0, 1],
    #                 [0, 2],
    #                 [1, 3],
    #                 [1, 4]]
    # and idx_anch_neg = [[0, 5],
    #                     [0, 6],
    #                     [1, 7]]
    #                     [1, 8]]
    # We want to get the following correspondence:
    #     0 -> [5, 6]
    #     1 -> [7, 8]
    # We know that idx_neg is sorted lexicographically. So, we can fint the number
    # of all unique anchors in idx_anch_neg, and use torch.split to get the desired
    # groups.
    _, neg_counts = torch.unique(idx_anch_neg[:, 0], return_counts=True, sorted=True)
    neg_groups = torch.split(idx_anch_neg[:, 1], neg_counts.tolist())

    # Now for a pair (anchor, positive)we have
    # neg_group[i], where i is the number of the anchor, with all
    # the negatives that come along with that anchor.
    # In order to get all possible triplets that contains (anchor, positive)
    # we just need to repeat it as much as we have negative elements
    # in the neg_groups[i].
    # For example above, we want to generate the following triplets:
    #
    # we need to repeat [0, 1] as much as there are | [0, 1, 5]  |
    # negatives that come along with anchor 0       | [0, 1, 6]  | and we repeat [5, 6] as much
    #                                                 [0, 2, 5]  | as there are anchor 0 in idx_anch_pos
    #                                                 [0, 2, 6]  |
    #                                                 [1, 3, 7]
    #                                                 [1, 3, 8]
    #                                                 [1, 4, 7]
    #                                                 [1, 4, 8]

    # Repeat the negatives as much as there are correspondent anchors.
    idx_neg = torch.cat(tuple(itemgetter(*idx_anch_pos[:, 0].cpu().tolist())(neg_groups)))

    # Repeat the positive pairs as much as there are correspondent netatives and get idx_anch, idx_pos.
    n_negs_in_group = [len(g) for g in neg_groups]
    n_repeats = torch.tensor(tuple(itemgetter(*idx_anch_pos[:, 0].cpu().numpy())(n_negs_in_group))).to(idx_anch_pos)
    idx_anch_pos = idx_anch_pos.repeat_interleave(n_repeats, dim=0)
    idx_anch, idx_pos = idx_anch_pos.T

    # Choose randomly the required number of triplets.
    random_idx = np.random.choice(len(idx_anch), min(len(idx_anch), max_out_triplets), replace=False)
    random_idx = torch.tensor(random_idx).to(device)
    idx_anch = idx_anch[random_idx]
    idx_pos = idx_pos[random_idx]
    idx_neg = idx_neg[random_idx]

    idx_anch = idx_anch.tolist()
    idx_pos = idx_pos.tolist()
    idx_neg = idx_neg.tolist()

    return idx_anch, idx_pos, idx_neg


def get_available_triplets_naive(labels: List[int], max_out_triplets: int = maxsize) -> TTripletsIds:
    """
    For each label, the function generates all possible positive and negative pairs of triplets.
    """
    num_labels = len(labels)

    triplets = []
    for label in set(labels):
        ids_pos_cur = find_value_ids(labels, label)
        ids_neg_cur = set(range(num_labels)) - set(ids_pos_cur)

        # (l0, l1, n) and (l1, l0, n) are 2 different triplets
        # and we want both of them
        pos_pairs = list(combinations(ids_pos_cur, r=2)) + list(combinations(ids_pos_cur[::-1], r=2))

        tri = [(a, p, n) for (a, p), n in product(pos_pairs, ids_neg_cur)]
        triplets.extend(tri)

    triplets = sample(triplets, min(len(triplets), max_out_triplets))
    ids_anchor, ids_pos, ids_neg = zip(*triplets)

    return list(ids_anchor), list(ids_pos), list(ids_neg)


__all__ = ["AllTripletsMiner", "get_available_triplets"]



import os
from typing import Any, Dict

from oml.const import TCfg
from oml.interfaces.loggers import IPipelineLogger
from oml.lightning.pipelines.logging import (
    ClearMLPipelineLogger,
    MLFlowPipelineLogger,
    NeptunePipelineLogger,
    TensorBoardPipelineLogger,
    WandBPipelineLogger,
)
from oml.utils.misc import dictconfig_to_dict

LOGGERS_REGISTRY = {
    "wandb": WandBPipelineLogger,
    "neptune": NeptunePipelineLogger,
    "tensorboard": TensorBoardPipelineLogger,
    "mlflow": MLFlowPipelineLogger,
    "clearml": ClearMLPipelineLogger,
}

CLOUD_TOKEN_NAMES = {"wandb": "WANDB_API_KEY", "neptune": "NEPTUNE_API_TOKEN"}
TOKEN_ERROR_MESSAGE = (
    "{} logger is specified in your config file, "
    "but <{}> is not provided as a global environment variable."
    "Please, execute `export {}=your_token` before running the script."
)


def get_logger(name: str, **kwargs: Dict[str, Any]) -> IPipelineLogger:
    if (name in CLOUD_TOKEN_NAMES) and (CLOUD_TOKEN_NAMES[name] not in os.environ):
        token_name = CLOUD_TOKEN_NAMES[name]
        raise ValueError(TOKEN_ERROR_MESSAGE.format(name.upper(), token_name, token_name))

    return LOGGERS_REGISTRY[name](**kwargs)  # type: ignore


def get_logger_by_cfg(cfg: TCfg) -> IPipelineLogger:
    cfg = dictconfig_to_dict(cfg)
    logger = get_logger(cfg["name"], **cfg["args"])
    return logger


__all__ = ["LOGGERS_REGISTRY", "get_logger", "get_logger_by_cfg"]



from collections import Counter
from typing import Iterator, List, Union

import numpy as np

from oml.interfaces.samplers import IBatchSampler
from oml.utils.misc import smart_sample


class BalanceSampler(IBatchSampler):
    """
    This sampler takes ``n_instances`` for each of the ``n_labels`` to form the batches.
    Thus, the batch size is ``n_instances x n_labels``. This type of sampling can be found
    in the classical Person Re-Id paper -
    `In Defense of the Triplet Loss for Person Re-Identification`_.

    .. _In Defense of the Triplet Loss for Person Re-Identification:
        https://arxiv.org/abs/1703.07737

    The strategy for the dataset with ``L`` unique labels is the following:

    - Select ``n_labels`` of ``L`` labels for the 1st batch

    - Select ``n_instances`` for each label for the 1st batch

    - Select ``n_labels`` of ``L - n_labels`` remaining labels for 2nd batch

    - Select ``n_instances`` instances for each label for the 2nd batch

    - ...

    - The epoch ends after ``L // n_labels``.

    Thus, in each epoch, all the labels will be selected once, but this
    does not mean that all the instances will be picked.

    Behavior in corner cases:

    - If some label does not contain ``n_instances``, a choice will be made with repetition.

    - If ``L % n_labels != 0`` then we drop the last batch.

    """

    def __init__(self, labels: Union[List[int], np.ndarray], n_labels: int, n_instances: int):
        """
        Args:
            labels: List of the labels for each element in the dataset
            n_labels: The desired number of labels in a batch, should be > 1
            n_instances: The desired number of instances of each label in a batch, should be > 1

        """
        unq_labels = set(labels)

        assert isinstance(n_labels, int) and isinstance(n_instances, int)
        assert (1 < n_labels <= len(unq_labels)) and (1 < n_instances)
        assert all(n > 1 for n in Counter(labels).values()), "Each label should contain at least 2 samples"

        self._labels = np.array(labels)
        self.n_labels = n_labels
        self.n_instances = n_instances

        self._batch_size = self.n_labels * self.n_instances
        self._unq_labels = unq_labels

        labels = np.array(labels)
        self.lbl2idx = {label: np.arange(len(labels))[labels == label].tolist() for label in set(labels)}

        self._batches_in_epoch = len(self._unq_labels) // self.n_labels

    @property
    def batch_size(self) -> int:
        return self._batch_size

    def __len__(self) -> int:
        return self._batches_in_epoch

    def __iter__(self) -> Iterator[List[int]]:
        inds_epoch = []

        labels_rest = self._unq_labels.copy()

        for _ in range(len(self)):
            ids_batch = []

            labels_for_batch = set(
                np.random.choice(list(labels_rest), size=min(self.n_labels, len(labels_rest)), replace=False)
            )
            labels_rest -= labels_for_batch

            for cls in labels_for_batch:
                cls_ids = self.lbl2idx[cls]
                selected_inds = smart_sample(cls_ids, self.n_instances)
                ids_batch.extend(selected_inds)

            inds_epoch.append(ids_batch)

        return iter(inds_epoch)  # type: ignore


__all__ = ["BalanceSampler"]



from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import torch
from torch import Tensor, nn
from torch.nn.modules.activation import Sigmoid
from torchvision.ops import MLP

from oml.interfaces.models import IExtractor, IFreezable, IPairwiseModel
from oml.models.utils import remove_prefix_from_state_dict
from oml.utils.io import download_checkpoint
from oml.utils.misc_torch import elementwise_dist


class LinearTrivialDistanceSiamese(IPairwiseModel):
    """
    This model is a useful tool mostly for development.

    """

    def __init__(self, feat_dim: int, identity_init: bool, output_bias: float = 0):
        """
        Args:
            feat_dim: Expected size of each input.
            identity_init: If ``True``, models' weights initialised in a way when
                the model simply estimates L2 distance between the original embeddings.
            output_bias: Value to add to the output.

        """
        super(LinearTrivialDistanceSiamese, self).__init__()
        self.feat_dim = feat_dim
        self.output_bias = output_bias

        self.proj = torch.nn.Linear(in_features=feat_dim, out_features=feat_dim, bias=False)

        if identity_init:
            self.proj.load_state_dict({"weight": torch.eye(feat_dim)})

    def forward(self, x1: Tensor, x2: Tensor) -> Tensor:
        """
        Args:
            x1: Embedding with the shape of ``[batch_size, feat_dim]``
            x2: Embedding with the shape of ``[batch_size, feat_dim]``

        Returns:
            Distance between transformed inputs.

        """
        x1 = self.proj(x1)
        x2 = self.proj(x2)
        y = elementwise_dist(x1, x2, p=2) + self.output_bias
        return y

    def predict(self, x1: Tensor, x2: Tensor) -> Tensor:
        return self.forward(x1=x1, x2=x2)


class ConcatSiamese(IPairwiseModel, IFreezable):
    """
    This model concatenates two inputs and passes them through
    a given backbone and applyies a head after that.

    """

    pretrained_models: Dict[str, Any] = {}

    def __init__(
        self,
        extractor: IExtractor,
        mlp_hidden_dims: List[int],
        use_tta: bool = False,
        weights: Optional[Union[str, Path]] = None,
    ) -> None:
        """
        Args:
            extractor: Instance of ``IExtractor`` (e.g. ``ViTExtractor``)
            mlp_hidden_dims: Hidden dimensions of the head
            use_tta: Set ``True`` if you want to average the results obtained by two different orders of concatenating
             input images. Affects only ``self.predict()`` method.
            weights: Path to weights file or ``None`` for random initialization

        """
        super(ConcatSiamese, self).__init__()
        self.extractor = extractor
        self.use_tta = use_tta

        self.head = MLP(
            in_channels=self.extractor.feat_dim,
            hidden_channels=[*mlp_hidden_dims, 1],
            activation_layer=Sigmoid,
            dropout=0.5,
            inplace=None,
        )

        # turn off the last bias
        self.head[-2] = nn.Linear(self.head[-2].in_features, self.head[-2].out_features, bias=False)

        # turn off the last dropout
        self.head[-1] = nn.Identity()

        self.train_backbone = True

        if weights:
            if weights in self.pretrained_models:
                url_or_fid, hash_md5, fname = self.pretrained_models[weights]  # type: ignore
                weights = download_checkpoint(url_or_fid=url_or_fid, hash_md5=hash_md5, fname=fname)

            loaded = torch.load(weights, map_location="cpu")
            loaded = loaded.get("state_dict", loaded)
            loaded = remove_prefix_from_state_dict(loaded, trial_key="extractor.")
            self.load_state_dict(loaded, strict=True)

    def forward(self, x1: Tensor, x2: Tensor) -> Tensor:
        x = torch.concat([x1, x2], dim=2)

        with torch.set_grad_enabled(self.train_backbone):
            x = self.extractor(x)

        x = self.head(x)
        x = x.view(len(x))

        return x

    def predict(self, x1: Tensor, x2: Tensor) -> Tensor:
        x = self.forward(x1=x1, x2=x2)
        x = torch.sigmoid(x)

        if self.use_tta:
            y = self.forward(x1=x2, x2=x1)
            y = torch.sigmoid(y)
            return (x + y) / 2
        else:
            return x

    def freeze(self) -> None:
        self.train_backbone = False

    def unfreeze(self) -> None:
        self.train_backbone = True


class TrivialDistanceSiamese(IPairwiseModel):
    """
    This model is a useful tool mostly for development.

    """

    pretrained_models: Dict[str, Any] = {}

    def __init__(self, extractor: IExtractor, output_bias: float = 0) -> None:
        """
        Args:
            extractor: Instance of ``IExtractor`` (e.g. ``ViTExtractor``)
            output_bias: Value to add to the outputs.

        """
        super(TrivialDistanceSiamese, self).__init__()
        self.extractor = extractor
        self.output_bias = output_bias

    def forward(self, x1: Tensor, x2: Tensor) -> Tensor:
        """
        Args:
            x1: The first input.
            x2: The second input.

        Returns:
            Distance between inputs.

        """
        x1 = self.extractor(x1)
        x2 = self.extractor(x2)
        return elementwise_dist(x1, x2, p=2) + self.output_bias

    def predict(self, x1: Tensor, x2: Tensor) -> Tensor:
        return self.forward(x1=x1, x2=x2)


__all__ = ["LinearTrivialDistanceSiamese", "ConcatSiamese", "TrivialDistanceSiamese"]



from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd
from torch import BoolTensor, LongTensor

from oml.const import (
    CATEGORIES_COLUMN,
    IS_GALLERY_COLUMN,
    IS_QUERY_COLUMN,
    LABELS_COLUMN,
    LABELS_KEY,
    SEQUENCE_COLUMN,
)
from oml.interfaces.datasets import (
    IBaseDataset,
    ILabeledDataset,
    IQueryGalleryDataset,
    IQueryGalleryLabeledDataset,
)


def update_extra_data(dataset: IBaseDataset, df: pd.DataFrame, extra_data: Dict[str, Any]) -> IBaseDataset:
    extra_data = dict() if extra_data is None else extra_data

    if CATEGORIES_COLUMN in df.columns:
        extra_data[CATEGORIES_COLUMN] = df[CATEGORIES_COLUMN].copy()

    if SEQUENCE_COLUMN in df.columns:
        extra_data[SEQUENCE_COLUMN] = df[SEQUENCE_COLUMN].copy()

    dataset.extra_data.update(extra_data)

    return dataset


def build_label2category(df: pd.DataFrame) -> Optional[Dict[int, Union[str, int]]]:
    if CATEGORIES_COLUMN in df.columns:
        label2category = dict(zip(df[LABELS_COLUMN], df[CATEGORIES_COLUMN]))
    else:
        label2category = None

    return label2category


class DFLabeledDataset(ILabeledDataset):
    def __init__(
        self,
        dataset: IBaseDataset,
        df: pd.DataFrame,
        extra_data: Optional[Dict[str, Any]] = None,
        labels_key: str = LABELS_KEY,
    ):
        assert LABELS_COLUMN in df.columns
        dataset = update_extra_data(dataset, df, extra_data)

        self._dataset = dataset
        self._labels = np.array(df[LABELS_COLUMN].copy())
        self._label2category = build_label2category(df)

        self.df = df
        self.extra_data = self._dataset.extra_data
        self.labels_key = labels_key
        self.index_key = self._dataset.index_key
        self.input_tensors_key = self._dataset.input_tensors_key

    def __len__(self) -> int:
        return len(self._dataset)

    def __getitem__(self, item: int) -> Dict[str, Any]:
        data = self._dataset[item]
        data[self.labels_key] = self._labels[item]
        return data

    def get_labels(self) -> np.ndarray:
        return self._labels

    def get_label2category(self) -> Optional[Dict[int, Union[str, int]]]:
        return self._label2category


class DFQueryGalleryDataset(IQueryGalleryDataset):
    def __init__(
        self,
        dataset: IBaseDataset,
        df: pd.DataFrame,
        extra_data: Optional[Dict[str, Any]] = None,
    ):
        assert all(x in df.columns for x in (IS_QUERY_COLUMN, IS_GALLERY_COLUMN))
        dataset = update_extra_data(dataset, df, extra_data)

        self._dataset = dataset
        self._query_ids = BoolTensor(df[IS_QUERY_COLUMN]).nonzero().squeeze()
        self._gallery_ids = BoolTensor(df[IS_GALLERY_COLUMN]).nonzero().squeeze()

        self.df = df
        self.extra_data = self._dataset.extra_data
        self.index_key = self._dataset.index_key
        self.input_tensors_key = self._dataset.input_tensors_key

    def __len__(self) -> int:
        return len(self._dataset)

    def __getitem__(self, item: int) -> Dict[str, Any]:
        return self._dataset[item]

    def get_query_ids(self) -> LongTensor:
        return self._query_ids

    def get_gallery_ids(self) -> LongTensor:
        return self._gallery_ids


class DFQueryGalleryLabeledDataset(IQueryGalleryLabeledDataset):
    def __init__(
        self,
        dataset: IBaseDataset,
        df: pd.DataFrame,
        extra_data: Optional[Dict[str, Any]] = None,
        labels_key: str = LABELS_KEY,
    ):
        assert all(x in df.columns for x in (IS_QUERY_COLUMN, IS_GALLERY_COLUMN, LABELS_COLUMN))
        dataset = update_extra_data(dataset, df, extra_data)

        self._dataset = dataset
        self._query_ids = BoolTensor(df[IS_QUERY_COLUMN]).nonzero().squeeze()
        self._gallery_ids = BoolTensor(df[IS_GALLERY_COLUMN]).nonzero().squeeze()
        self._labels = np.array(df[LABELS_COLUMN].copy())
        self._label2category = build_label2category(df)

        self.df = df
        self.extra_data = self._dataset.extra_data
        self.index_key = self._dataset.index_key
        self.input_tensors_key = self._dataset.input_tensors_key
        self.labels_key = labels_key

    def __len__(self) -> int:
        return len(self._dataset)

    def __getitem__(self, item: int) -> Dict[str, Any]:
        data = self._dataset[item]
        data[self.labels_key] = self._labels[item]
        return data

    def get_query_ids(self) -> LongTensor:
        return self._query_ids

    def get_gallery_ids(self) -> LongTensor:
        return self._gallery_ids

    def get_labels(self) -> np.ndarray:
        return self._labels

    def get_label2category(self) -> Optional[Dict[int, Union[str, int]]]:
        return self._label2category


__all__ = ["DFLabeledDataset", "DFQueryGalleryDataset", "DFQueryGalleryLabeledDataset"]



import warnings
from math import ceil
from typing import Any, Optional

import matplotlib.pyplot as plt
import pytorch_lightning as pl
from pytorch_lightning import Callback
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.utils.data import DataLoader

from oml.const import INDEX_KEY, LOG_IMAGE_FOLDER
from oml.ddp.patching import check_loaders_is_patched, patch_dataloader_to_ddp
from oml.interfaces.loggers import IFigureLogger
from oml.interfaces.metrics import IBasicMetric, IMetricVisualisable
from oml.lightning.modules.ddp import ModuleDDP
from oml.utils.misc import flatten_dict


class MetricValCallback(Callback):
    """
    This is a wrapper which allows to use IBasicMetric with PyTorch Lightning.

    """

    def __init__(
        self,
        metric: IBasicMetric,
        log_images: bool = False,
        loader_idx: int = 0,
    ):
        """
        Args:
            metric: Metric
            log_images: Set ``True`` if you want to have visual logging
            loader_idx: Idx of the loader to calculate metric for

        """

        self.metric = metric
        self.log_images = log_images
        assert not log_images or (isinstance(metric, IMetricVisualisable) and metric.ready_to_visualize())

        self.loader_idx = loader_idx

        self._expected_samples = 0
        self._collected_samples = 0
        self._ready_to_accumulate = False

    def _calc_expected_samples(self, trainer: pl.Trainer, dataloader_idx: int = 0) -> int:
        loaders = (
            [trainer.val_dataloaders] if isinstance(trainer.val_dataloaders, DataLoader) else trainer.val_dataloaders
        )
        len_dataset = len(loaders[dataloader_idx].dataset)
        if trainer.world_size > 1:
            # we use padding in DDP and sequential sampler for validation
            len_dataset = ceil(len_dataset / trainer.world_size)
        return len_dataset

    def on_validation_batch_start(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule, batch: Any, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        if dataloader_idx == self.loader_idx:
            if not self._ready_to_accumulate:
                self._expected_samples = self._calc_expected_samples(trainer=trainer, dataloader_idx=dataloader_idx)
                self._collected_samples = 0

                self.metric.setup(num_samples=self._expected_samples)
                self._ready_to_accumulate = True

    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        outputs: Optional[STEP_OUTPUT],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        if dataloader_idx == self.loader_idx:
            assert self._ready_to_accumulate

            self.metric.update_data(data=outputs, indices=outputs[INDEX_KEY])  # type: ignore

            self._collected_samples += len(outputs[INDEX_KEY])
            if self._collected_samples > self._expected_samples:
                self._raise_computation_error()

            # For non-additive metrics (like f1) we usually accumulate some information during the epoch, then we
            # calculate final score `on_validation_epoch_end`. The problem here is that `on_validation_epoch_end`
            # lightning logger doesn't know dataloader_idx and logs metrics in incorrect way if num_dataloaders > 1.
            # To avoid the problem we calculate metrics `on_validation_batch_end` for the last batch in the loader.
            is_last_expected_batch = self._collected_samples == self._expected_samples
            if is_last_expected_batch:
                self.calc_and_log_metrics(pl_module)

    def _log_images(self, pl_module: pl.LightningModule) -> None:
        if not isinstance(self.metric, IMetricVisualisable):
            return

        if not isinstance(pl_module.logger, IFigureLogger):
            warnings.warn(
                f"Unexpected logger {pl_module.logger}. Figures have not been saved. "
                f"Please, use a child of {IFigureLogger}."
            )
            return

        for fig, metric_log_str in zip(*self.metric.visualize()):
            log_str = f"{LOG_IMAGE_FOLDER}/{metric_log_str}"
            pl_module.logger.log_figure(fig=fig, title=log_str, idx=pl_module.current_epoch)
            plt.close(fig=fig)

    def on_validation_epoch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        self._ready_to_accumulate = False
        if self._collected_samples != self._expected_samples:
            self._raise_computation_error()

    def calc_and_log_metrics(self, pl_module: pl.LightningModule) -> None:
        metrics = self.metric.compute_metrics()
        metrics = flatten_dict(metrics)
        pl_module.log_dict(metrics, rank_zero_only=True, add_dataloader_idx=True)

        if self.log_images:
            self._log_images(pl_module=pl_module)

    def _raise_computation_error(self) -> Exception:
        raise ValueError(
            f"Incorrect calculation for {self.metric.__class__.__name__} metric. "
            f"Inconsistent number of samples, obtained: {self._collected_samples}, "
            f"expected: {self._expected_samples}. "
            f"Make sure that you don't use the 'overfit_batches' parameter in 'pl.Trainer' and "
            f"you set 'drop_last=False'. The idea is that lengths of dataset and dataloader must match."
        )

    @staticmethod
    def _check_loaders(trainer: "pl.Trainer") -> None:
        if trainer.world_size > 1 and trainer.val_dataloaders is not None:
            if not check_loaders_is_patched(trainer.val_dataloaders):
                raise RuntimeError(err_message_loaders_is_not_patched)

    def on_train_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._check_loaders(trainer)

    def on_validation_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        self._check_loaders(trainer)


err_message_loaders_is_not_patched = (
    "\nExperiment is runned in DDP mode, but some of validation dataloaders is not patched. Metric callback will "
    "be incorrect  without patched loaders. Possible problems and solutions:\n"
    f"1) If you use custom module inherited from 'pl.LightningModule', please replace ancestor class with "
    f"our '{ModuleDDP.__name__}', which automaticaly patches your loaders\n"
    f"2) If you implement your own 'train_dataloader' or 'val_dataloader' methods for your module, you can add "
    f"extra line of code for patching loader with '{patch_dataloader_to_ddp.__name__}' function\n"
    f"3) If you call 'trainer.fit(...)' or 'trainer.validate(...)' method with loaders as argument, PytorchLightning "
    f"will ignore loaders from 'train_dataloader' and 'val_dataloader' methods. Please avoid substituting loaders to "
    f"this functions, instead use '{ModuleDDP.__name__}'\n"
    f"4) Check that the flag 'replace_sampler_ddp=False' in the trainer constructor, because we do this "
    f"replacement in '{ModuleDDP.__name__}' constructor\n"
    f"5) Turn off the 'overfit_batches' parameter in 'pl.Trainer'."
)

__all__ = ["MetricValCallback"]



from typing import Iterable, OrderedDict, Union

import torch
from torch import nn

TStateDict = OrderedDict[str, torch.Tensor]


def remove_criterion_in_state_dict(state_dict: TStateDict) -> TStateDict:
    if "criterion.weight" in state_dict:
        del state_dict["criterion.weight"]
    return state_dict


def find_prefix_in_state_dict(state_dict: TStateDict, trial_key: str) -> str:
    keys_starting_with_trial_key = [k for k in state_dict.keys() if trial_key in k]
    assert keys_starting_with_trial_key, (
        f"There are no keys starting from {trial_key}.\n" f"The existing keys are: {list(state_dict.keys())}"
    )

    k0 = keys_starting_with_trial_key[0]
    prefix = k0[: k0.index(trial_key)]

    keys_not_starting_with_prefix = list(filter(lambda x: not x.startswith(prefix), state_dict.keys()))
    assert (
        not keys_not_starting_with_prefix
    ), f"There are keys not starting from the found prefix {prefix}: {keys_not_starting_with_prefix}"

    return prefix


def remove_prefix_from_state_dict(state_dict: TStateDict, trial_key: str) -> TStateDict:
    prefix = find_prefix_in_state_dict(state_dict, trial_key)

    if prefix == "":
        return state_dict

    else:

        for k in list(state_dict.keys()):
            if k.startswith(prefix):
                state_dict[k[len(prefix) :]] = state_dict[k]
                del state_dict[k]

        print(f"Prefix <{prefix}> was removed from the state dict.")

        return state_dict


def filter_state_dict(state_dict: TStateDict, needed_keys: Iterable[str]) -> TStateDict:

    for k in list(state_dict):
        if k not in needed_keys:
            del state_dict[k]

    return state_dict


def patch_float(module: nn.Module, float_node: torch.Node) -> None:
    """
    This function is for patching jitted weights with hardcoded ``.to(dtype)`` operation.
    """
    try:
        graphs = [module.graph] if hasattr(module, "graph") else []
    except RuntimeError:
        graphs = []

    if hasattr(module, "forward1"):
        graphs.append(module.forward1.graph)

    for graph in graphs:
        for node in graph.findAllNodes("aten::to"):
            inputs = list(node.inputs())
            for i in [1, 2]:  # dtype can be the second or third argument to aten::to()
                if "[value=5]" in repr(inputs[i]):
                    inputs[i].node().copyAttributes(float_node)

    for child in module.children():
        patch_float(child, float_node)


def patch_device(module: nn.Module, device_node: torch.Node) -> None:
    """
    This function is for patching jitted weights with hardcoded ``.to(device)`` operation.
    """
    try:
        graphs = [module.graph] if hasattr(module, "graph") else []
    except RuntimeError:
        graphs = []

    if hasattr(module, "forward1"):
        graphs.append(module.forward1.graph)

    for g in graphs:
        for node in g.findAllNodes("prim::Constant"):
            if "value" in node.attributeNames() and ('value="cuda' in repr(node)):
                node.copyAttributes(device_node)

    for child in module.children():
        patch_device(child, device_node)


def patch_device_and_float(module: nn.Module, device: Union[str, torch.device] = "cuda") -> None:
    """
    This function is for patching jitted weights with hardcoded ``.to(device)`` and ``.to(dtype)`` operations.
    You may need this if you want to correctly load some jitted model which uses half-precision and(or) which
    device was hardcoded.
    """
    device_holder = torch.jit.trace(lambda: torch.ones([]).to(torch.device(device)), example_inputs=[])
    device_node = [n for n in device_holder.graph.findAllNodes("prim::Constant") if "Device" in repr(n)][-1]
    patch_device(module, device_node)

    # patch dtype to float32 on CPU
    if str(device) == "cpu":
        float_holder = torch.jit.trace(lambda: torch.ones([]).float(), example_inputs=[])
        float_node = list(float_holder.graph.findNode("aten::to").inputs())[1].node()
        patch_float(module, float_node)


__all__ = [
    "find_prefix_in_state_dict",
    "remove_prefix_from_state_dict",
    "filter_state_dict",
    "patch_float",
    "patch_device",
    "patch_device_and_float",
]



# References:
#   https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/config.py

import os

import torch

# use torch.scaled_dot_product_attention where possible
_HAS_FUSED_ATTN = hasattr(torch.nn.functional, "scaled_dot_product_attention")
_USE_FUSED_ATTN = int(os.environ.get("USE_FUSED_ATTN", 0))


def use_fused_attn() -> bool:
    # NOTE: ONNX export cannot handle F.scaled_dot_product_attention as of pytorch 2.0
    if not _HAS_FUSED_ATTN:
        return False
    return _USE_FUSED_ATTN > 0



from typing import Any, Dict, List

from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    CyclicLR,
    ExponentialLR,
    LambdaLR,
    MultiplicativeLR,
    MultiStepLR,
    OneCycleLR,
    ReduceLROnPlateau,
    StepLR,
    _LRScheduler,
)

from oml.utils.misc import TCfg, dictconfig_to_dict


class ReduceLROnPlateauPatch(ReduceLROnPlateau):
    def get_last_lr(self) -> List[float]:
        return [group["lr"] for group in self.optimizer.param_groups]


SCHEDULERS_REGISTRY = {
    "lambda": LambdaLR,
    "multiplicative": MultiplicativeLR,
    "step": StepLR,
    "multi_step": MultiStepLR,
    "exponential": ExponentialLR,
    "cosine_annealing": CosineAnnealingLR,
    "cyclic": CyclicLR,
    "one_cycle": OneCycleLR,
    "reduce_on_plateau": ReduceLROnPlateauPatch,
}


def get_scheduler(name: str, **kwargs: Dict[str, Any]) -> _LRScheduler:
    return SCHEDULERS_REGISTRY[name](**kwargs)


def get_scheduler_by_cfg(cfg: TCfg, **kwargs_runtime: Dict[str, Any]) -> _LRScheduler:
    cfg = dictconfig_to_dict(cfg)
    cfg["args"].update(kwargs_runtime)
    return get_scheduler(name=cfg["name"], **cfg["args"])


__all__ = ["SCHEDULERS_REGISTRY", "get_scheduler", "get_scheduler_by_cfg"]



# type: ignore
# flake8: noqa


# References:
#   https://github.com/facebookresearch/dinov2/blob/main/dinov2/models/vision_transformer.py
#   https://github.com/rwightman/pytorch-image-models/tree/master/timm/models/vision_transformer.py


import math
from functools import partial
from typing import Callable, Optional, Sequence, Tuple, Union

import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch.nn.init import trunc_normal_

from oml.models.vit_dino.external_v2.config import use_fused_attn


def named_apply(fn: Callable, module: nn.Module, name="", depth_first=True, include_root=False) -> nn.Module:
    if not depth_first and include_root:
        fn(module=module, name=name)
    for child_name, child_module in module.named_children():
        child_name = ".".join((name, child_name)) if name else child_name
        named_apply(
            fn=fn,
            module=child_module,
            name=child_name,
            depth_first=depth_first,
            include_root=True,
        )
    if depth_first and include_root:
        fn(module=module, name=name)
    return module


def init_weights_vit_timm(module: nn.Module, name: str = ""):
    """ViT weight initialization, original timm impl (for reproducibility)"""
    if isinstance(module, nn.Linear):
        trunc_normal_(module.weight, std=0.02)
        if module.bias is not None:
            nn.init.zeros_(module.bias)


def make_2tuple(x):
    if isinstance(x, tuple):
        assert len(x) == 2
        return x

    assert isinstance(x, int)
    return (x, x)


def drop_path(x, drop_prob: float = 0.0, training: bool = False):
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
    if keep_prob > 0.0:
        random_tensor.div_(keep_prob)
    output = x * random_tensor
    return output


def drop_add_residual_stochastic_depth(
    x: Tensor,
    residual_func: Callable[[Tensor], Tensor],
    sample_drop_ratio: float = 0.0,
) -> Tensor:
    # 1) extract subset using permutation
    b, n, d = x.shape
    sample_subset_size = max(int(b * (1 - sample_drop_ratio)), 1)
    brange = (torch.randperm(b, device=x.device))[:sample_subset_size]
    x_subset = x[brange]

    # 2) apply residual_func to get residual
    residual = residual_func(x_subset)

    x_flat = x.flatten(1)
    residual = residual.flatten(1)

    residual_scale_factor = b / sample_subset_size

    # 3) add the residual
    x_plus_residual = torch.index_add(x_flat, 0, brange, residual.to(dtype=x.dtype), alpha=residual_scale_factor)
    return x_plus_residual.view_as(x)


class BlockChunk(nn.ModuleList):
    def forward(self, x):
        for b in self:
            x = b(x)
        return x


class DropPath(nn.Module):
    def __init__(self, drop_prob: float = 0.0):
        super().__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)


class LayerScale(nn.Module):
    def __init__(
        self,
        dim: int,
        init_values: Union[float, Tensor] = 1e-5,
        inplace: bool = False,
    ) -> None:
        super().__init__()
        self.inplace = inplace
        self.gamma = nn.Parameter(init_values * torch.ones(dim))

    def forward(self, x: Tensor) -> Tensor:
        return x.mul_(self.gamma) if self.inplace else x * self.gamma


class Attention(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int = 8,
        qkv_bias: bool = False,
        proj_bias: bool = True,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
    ) -> None:
        super().__init__()
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim**-0.5
        self.fused_attn = use_fused_attn()

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim, bias=proj_bias)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)

        if self.fused_attn:
            x = F.scaled_dot_product_attention(
                q,
                k,
                v,
                dropout_p=self.attn_drop.p if self.training else 0.0,
            )
        else:
            q = q * self.scale
            attn = q @ k.transpose(-2, -1)
            attn = attn.softmax(dim=-1)
            attn = self.attn_drop(attn)
            x = attn @ v

        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class Mlp(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: Optional[int] = None,
        out_features: Optional[int] = None,
        act_layer: Callable[..., nn.Module] = nn.GELU,
        drop: float = 0.0,
        bias: bool = True,
    ) -> None:
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features, bias=bias)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features, bias=bias)
        self.drop = nn.Dropout(drop)

    def forward(self, x: Tensor) -> Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class PatchEmbed(nn.Module):
    """
    2D image to patch embedding: (B,C,H,W) -> (B,N,D)

    Args:
        img_size: Image size.
        patch_size: Patch token size.
        in_chans: Number of input image channels.
        embed_dim: Number of linear projection output channels.
        norm_layer: Normalization layer.
    """

    def __init__(
        self,
        img_size: Union[int, Tuple[int, int]] = 224,
        patch_size: Union[int, Tuple[int, int]] = 16,
        in_chans: int = 3,
        embed_dim: int = 768,
        norm_layer: Optional[Callable] = None,
        flatten_embedding: bool = True,
    ) -> None:
        super().__init__()

        image_HW = make_2tuple(img_size)
        patch_HW = make_2tuple(patch_size)
        patch_grid_size = (
            image_HW[0] // patch_HW[0],
            image_HW[1] // patch_HW[1],
        )

        self.img_size = image_HW
        self.patch_size = patch_HW
        self.patches_resolution = patch_grid_size
        self.num_patches = patch_grid_size[0] * patch_grid_size[1]

        self.in_chans = in_chans
        self.embed_dim = embed_dim

        self.flatten_embedding = flatten_embedding

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_HW, stride=patch_HW)
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        _, _, H, W = x.shape
        patch_H, patch_W = self.patch_size

        assert H % patch_H == 0, f"Input image height {H} is not a multiple of patch height {patch_H}"
        assert W % patch_W == 0, f"Input image width {W} is not a multiple of patch width: {patch_W}"

        x = self.proj(x)  # B C H W
        H, W = x.size(2), x.size(3)
        x = x.flatten(2).transpose(1, 2)  # B HW C
        x = self.norm(x)
        if not self.flatten_embedding:
            x = x.reshape(-1, H, W, self.embed_dim)  # B H W C
        return x


class Block(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = False,
        proj_bias: bool = True,
        ffn_bias: bool = True,
        drop: float = 0.0,
        attn_drop: float = 0.0,
        init_values=None,
        drop_path: float = 0.0,
        act_layer: Callable[..., nn.Module] = nn.GELU,
        norm_layer: Callable[..., nn.Module] = nn.LayerNorm,
        attn_class: Callable[..., nn.Module] = Attention,
        ffn_layer: Callable[..., nn.Module] = Mlp,
    ) -> None:
        super().__init__()
        # print(f"biases: qkv: {qkv_bias}, proj: {proj_bias}, ffn: {ffn_bias}")
        self.norm1 = norm_layer(dim)
        self.attn = attn_class(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            proj_bias=proj_bias,
            attn_drop=attn_drop,
            proj_drop=drop,
        )
        self.ls1 = LayerScale(dim, init_values=init_values) if init_values else nn.Identity()
        self.drop_path1 = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = ffn_layer(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
            bias=ffn_bias,
        )
        self.ls2 = LayerScale(dim, init_values=init_values) if init_values else nn.Identity()
        self.drop_path2 = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.sample_drop_ratio = drop_path

    def forward(self, x: Tensor) -> Tensor:
        def attn_residual_func(x: Tensor) -> Tensor:
            return self.ls1(self.attn(self.norm1(x)))

        def ffn_residual_func(x: Tensor) -> Tensor:
            return self.ls2(self.mlp(self.norm2(x)))

        if self.training and self.sample_drop_ratio > 0.1:
            # the overhead is compensated only for a drop path rate larger than 0.1
            x = drop_add_residual_stochastic_depth(
                x,
                residual_func=attn_residual_func,
                sample_drop_ratio=self.sample_drop_ratio,
            )
            x = drop_add_residual_stochastic_depth(
                x,
                residual_func=ffn_residual_func,
                sample_drop_ratio=self.sample_drop_ratio,
            )
        elif self.training and self.sample_drop_ratio > 0.0:
            x = x + self.drop_path1(attn_residual_func(x))
            x = x + self.drop_path2(ffn_residual_func(x))
        else:
            x = x + attn_residual_func(x)
            x = x + ffn_residual_func(x)
        return x


class DinoVisionTransformer(nn.Module):
    def __init__(
        self,
        img_size: Union[int, Tuple[int, int]] = 224,
        patch_size: int = 16,
        in_chans: int = 3,
        embed_dim: int = 768,
        depth: int = 12,
        num_heads: int = 12,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        ffn_bias: bool = True,
        proj_bias: bool = True,
        drop_path_rate: float = 0.0,
        drop_path_uniform: bool = False,
        init_values: Optional[float] = None,  # for layerscale: None or 0 => no layerscale
        embed_layer: Callable[..., nn.Module] = PatchEmbed,
        act_layer: Callable[..., nn.Module] = nn.GELU,
        block_fn: Callable[..., nn.Module] = Block,
        ffn_layer="mlp",
        block_chunks: int = 1,
        num_register_tokens: int = 0,
        interpolate_antialias: bool = False,
        interpolate_offset: float = 0.1,
    ):
        super().__init__()
        norm_layer = partial(nn.LayerNorm, eps=1e-6)

        self.num_features = self.embed_dim = embed_dim  # num_features for consistency with other models
        self.num_tokens = 1
        self.n_blocks = depth
        self.num_heads = num_heads
        self.patch_size = patch_size
        self.num_register_tokens = num_register_tokens
        self.interpolate_antialias = interpolate_antialias
        self.interpolate_offset = interpolate_offset

        self.patch_embed = embed_layer(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=embed_dim,
        )
        num_patches = self.patch_embed.num_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + self.num_tokens, embed_dim))
        assert num_register_tokens >= 0
        self.register_tokens = (
            nn.Parameter(torch.zeros(1, num_register_tokens, embed_dim)) if num_register_tokens else None
        )

        if drop_path_uniform is True:
            dpr = [drop_path_rate] * depth
        else:
            dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule

        if ffn_layer == "mlp":
            ffn_layer = Mlp
        else:
            raise NotImplementedError

        blocks_list = [
            block_fn(
                dim=embed_dim,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                proj_bias=proj_bias,
                ffn_bias=ffn_bias,
                drop_path=dpr[i],
                norm_layer=norm_layer,
                act_layer=act_layer,
                ffn_layer=ffn_layer,
                init_values=init_values,
            )
            for i in range(depth)
        ]
        if block_chunks > 0:
            self.chunked_blocks = True
            chunked_blocks = []
            chunksize = depth // block_chunks
            for i in range(0, depth, chunksize):
                # this is to keep the block index consistent if we chunk the block list
                chunked_blocks.append([nn.Identity()] * i + blocks_list[i : i + chunksize])
            self.blocks = nn.ModuleList([BlockChunk(p) for p in chunked_blocks])
        else:
            self.chunked_blocks = False
            self.blocks = nn.ModuleList(blocks_list)

        self.norm = norm_layer(embed_dim)
        self.head = nn.Identity()

        self.mask_token = nn.Parameter(torch.zeros(1, embed_dim))

        self.init_weights()

    def init_weights(self):
        trunc_normal_(self.pos_embed, std=0.02)
        nn.init.normal_(self.cls_token, std=1e-6)
        if self.register_tokens is not None:
            nn.init.normal_(self.register_tokens, std=1e-6)
        named_apply(init_weights_vit_timm, self)

    def interpolate_pos_encoding(self, x, w, h):
        previous_dtype = x.dtype
        npatch = x.shape[1] - 1
        N = self.pos_embed.shape[1] - 1
        if npatch == N and w == h:
            return self.pos_embed
        pos_embed = self.pos_embed.float()
        class_pos_embed = pos_embed[:, 0]
        patch_pos_embed = pos_embed[:, 1:]
        dim = x.shape[-1]
        w0 = w // self.patch_size
        h0 = h // self.patch_size
        M = int(math.sqrt(N))  # Recover the number of patches in each dimension
        assert N == M * M
        kwargs = {}
        if self.interpolate_offset:
            sx = float(w0 + self.interpolate_offset) / M
            sy = float(h0 + self.interpolate_offset) / M
            kwargs["scale_factor"] = (sx, sy)
        else:
            # Simply specify an output size instead of a scale factor
            kwargs["size"] = (w0, h0)
        patch_pos_embed = nn.functional.interpolate(
            patch_pos_embed.reshape(1, M, M, dim).permute(0, 3, 1, 2),
            mode="bicubic",
            antialias=self.interpolate_antialias,
            **kwargs,
        )
        assert (w0, h0) == patch_pos_embed.shape[-2:]
        patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
        return torch.cat((class_pos_embed.unsqueeze(0), patch_pos_embed), dim=1).to(previous_dtype)

    def prepare_tokens_with_masks(self, x, masks=None):
        B, nc, w, h = x.shape
        x = self.patch_embed(x)
        if masks is not None:
            x = torch.where(masks.unsqueeze(-1), self.mask_token.to(x.dtype).unsqueeze(0), x)

        x = torch.cat((self.cls_token.expand(x.shape[0], -1, -1), x), dim=1)
        x = x + self.interpolate_pos_encoding(x, w, h)

        if self.register_tokens is not None:
            x = torch.cat(
                (
                    x[:, :1],
                    self.register_tokens.expand(x.shape[0], -1, -1),
                    x[:, 1:],
                ),
                dim=1,
            )

        return x

    def forward_features_list(self, x_list, masks_list):
        x = [self.prepare_tokens_with_masks(x, masks) for x, masks in zip(x_list, masks_list)]
        for blk in self.blocks:
            x = blk(x)

        all_x = x
        output = []
        for x, masks in zip(all_x, masks_list):
            x_norm = self.norm(x)
            output.append(
                {
                    "x_norm_clstoken": x_norm[:, 0],
                    "x_norm_regtokens": x_norm[:, 1 : self.num_register_tokens + 1],
                    "x_norm_patchtokens": x_norm[:, self.num_register_tokens + 1 :],
                    "x_prenorm": x,
                    "masks": masks,
                }
            )
        return output

    def forward_features(self, x, masks=None):
        if isinstance(x, list):
            return self.forward_features_list(x, masks)

        x = self.prepare_tokens_with_masks(x, masks)

        for blk in self.blocks:
            x = blk(x)

        x_norm = self.norm(x)
        return {
            "x_norm_clstoken": x_norm[:, 0],
            "x_norm_regtokens": x_norm[:, 1 : self.num_register_tokens + 1],
            "x_norm_patchtokens": x_norm[:, self.num_register_tokens + 1 :],
            "x_prenorm": x,
            "masks": masks,
        }

    def _get_intermediate_layers_not_chunked(self, x, n: Union[int, Sequence] = 1):
        x = self.prepare_tokens_with_masks(x)
        # If n is an int, take the n last blocks. If it's a list, take them
        output, total_block_len = [], len(self.blocks)
        blocks_to_take = range(total_block_len - n, total_block_len) if isinstance(n, int) else n
        for i, blk in enumerate(self.blocks):
            x = blk(x)
            if i in blocks_to_take:
                output.append(x)
        assert len(output) == len(blocks_to_take), f"only {len(output)} / {len(blocks_to_take)} blocks found"
        return output

    def _get_intermediate_layers_chunked(self, x, n: Union[int, Sequence] = 1):
        x = self.prepare_tokens_with_masks(x)
        output, i, total_block_len = [], 0, len(self.blocks[-1])  # type: ignore
        # If n is an int, take the n last blocks. If it's a list, take them
        blocks_to_take = range(total_block_len - n, total_block_len) if isinstance(n, int) else n
        for block_chunk in self.blocks:
            for blk in block_chunk[i:]:  # type: ignore
                x = blk(x)
                if i in blocks_to_take:
                    output.append(x)
                i += 1
        assert len(output) == len(blocks_to_take), f"only {len(output)} / {len(blocks_to_take)} blocks found"
        return output

    def get_intermediate_layers(
        self,
        x: torch.Tensor,
        n: Union[int, Sequence] = 1,  # Layers or n last layers to take
        reshape: bool = False,
        return_class_token: bool = False,
        norm=True,
    ) -> Tuple[Union[torch.Tensor, Tuple[torch.Tensor]]]:
        if self.chunked_blocks:
            outputs = self._get_intermediate_layers_chunked(x, n)
        else:
            outputs = self._get_intermediate_layers_not_chunked(x, n)
        if norm:
            outputs = [self.norm(out) for out in outputs]
        class_tokens = [out[:, 0] for out in outputs]
        outputs = [out[:, 1 + self.num_register_tokens :] for out in outputs]
        if reshape:
            B, _, w, h = x.shape
            outputs = [
                out.reshape(B, w // self.patch_size, h // self.patch_size, -1).permute(0, 3, 1, 2).contiguous()
                for out in outputs
            ]
        if return_class_token:
            return tuple(zip(outputs, class_tokens))  # type: ignore
        return tuple(outputs)

    def forward(self, *args, is_training=False, **kwargs):
        ret = self.forward_features(*args, **kwargs)
        if is_training:
            return ret
        else:
            return self.head(ret["x_norm_clstoken"])  # type: ignore


def vit_small(patch_size=16, num_register_tokens=0, **kwargs):
    model = DinoVisionTransformer(
        patch_size=patch_size,
        embed_dim=384,
        depth=12,
        num_heads=6,
        mlp_ratio=4,
        block_fn=partial(Block, attn_class=Attention),
        num_register_tokens=num_register_tokens,
        **kwargs,
    )
    return model


def vit_base(patch_size=16, num_register_tokens=0, **kwargs):
    model = DinoVisionTransformer(
        patch_size=patch_size,
        embed_dim=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4,
        block_fn=partial(Block, attn_class=Attention),
        num_register_tokens=num_register_tokens,
        **kwargs,
    )
    return model


def vit_large(patch_size=16, num_register_tokens=0, **kwargs):
    model = DinoVisionTransformer(
        patch_size=patch_size,
        embed_dim=1024,
        depth=24,
        num_heads=16,
        mlp_ratio=4,
        block_fn=partial(Block, attn_class=Attention),
        num_register_tokens=num_register_tokens,
        **kwargs,
    )
    return model



from random import choice
from typing import Dict, Optional, Tuple

from torch import (
    Tensor,
    abs,
    arange,
    cartesian_prod,
    cat,
    clip,
    combinations,
    flip,
    no_grad,
    randint,
    randperm,
    tensor,
    unique,
    zeros,
)

from oml.interfaces.miners import ITripletsMiner


class TripletMinerWithMemory(ITripletsMiner):
    """
    This miner has a memory bank that allows to sample not only the triplets from the original batch,
    but also add batches obtained from both the bank and the original batch.

    """

    def __init__(self, bank_size_in_batches: int, tri_expand_k: int):
        """

        Args:
            bank_size_in_batches: The size of the bank calculated in the number batches
            tri_expand_k: This parameter defines how many triplets we sample from the bank.
                 Specifically, we return ``tri_expand_k * number of original triplets``.
                 In particular, if ``tri_expand_k == 1`` we sample no triplets from the bank

        """
        assert tri_expand_k >= 1

        self.bank_size_in_batches = bank_size_in_batches
        self.tri_expand_k = tri_expand_k

        self.bank_features: Optional[Tensor] = None
        self.bank_labels: Optional[Tensor] = None
        self.bs = -1
        self.bank_size = -1
        self.ptr = 0

    @no_grad()
    def __allocate_if_needed(self, features: Tensor, labels: Tensor) -> None:
        if self.bank_features is None:
            assert len(features) == len(labels)

            self.bs = features.shape[0]
            self.feat_dim = features.shape[-1]
            self.bank_size = self.bank_size_in_batches * self.bs

            # let's init our bank with the following labels: -1, -2, -3, -4 ...
            # and use one-hot encoding for their features
            self.bank_labels = -1 * arange(1, self.bs + 1).repeat(self.bank_size_in_batches).long()
            self.bank_features = zeros([self.bank_size, self.feat_dim], dtype=features.dtype).to(features.device)
            self.bank_features[arange(self.bank_size), clip(abs(self.bank_labels), max=self.feat_dim - 1)] = 1

    @no_grad()
    def update_bank(self, features: Tensor, labels: Tensor) -> None:
        self.bank_features[self.ptr : self.ptr + self.bs] = features.clone().detach()
        self.bank_labels[self.ptr : self.ptr + self.bs] = labels.clone()
        self.ptr = (self.ptr + self.bs) % self.bank_size

    @no_grad()
    def get_pos_pairs(self, lbl2idx: Dict[Tensor, Tensor], n: int = None) -> Tensor:
        pos_batch_pairs = zeros(0, 2)

        if n is not None:
            while len(pos_batch_pairs) < n:
                pos_ii = choice(list(lbl2idx.values()))
                combs = combinations(pos_ii, r=2)
                pos_batch_pairs = cat([pos_batch_pairs, combs, flip(combs, [1])])
        else:
            for pos_ii in lbl2idx.values():
                combs = combinations(pos_ii, r=2)
                pos_batch_pairs = cat([pos_batch_pairs, combs, flip(combs, [1])])

        return pos_batch_pairs.long()[randperm(len(pos_batch_pairs))[:n]]

    def sample(  # type: ignore
        self, features: Tensor, labels: Tensor  # type: ignore
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:  # type: ignore
        """

        Args:
            features: Features with the shape of ``(batch_size, feat_dim)``
            labels: Labels with the size of ``batch_size``

        Returns:
            Triplets made from the original batch and those that were combined from the bank and the batch.
            We also return an indicator of whether triplet was obtained from the original batch.
            So, output is the following ``(anchor, positive, negative, indicators)``

        """
        labels = tensor(labels).long()
        self.__allocate_if_needed(features=features, labels=labels)

        assert len(features) == len(labels) == self.bs, (len(features), len(labels), self.bs)

        # todo: optimize performance
        lbl2idx_bank = {lb: arange(self.bank_size)[self.bank_labels == lb] for lb in unique(self.bank_labels)}
        lbl2idx_batch = {lb: arange(self.bs)[labels == lb] for lb in unique(labels)}

        # part1: anchor + positive + negative come from batch
        ii_anch_pos_1 = self.get_pos_pairs(lbl2idx_batch)
        ii_all = arange(self.bs)
        ii_pos_pairs_1, ii_neg_1 = cartesian_prod(arange(len(ii_anch_pos_1)), ii_all).T
        ii_anch_1, ii_pos_1 = ii_anch_pos_1[ii_pos_pairs_1].T

        ii_anch_1, ii_pos_1, ii_neg_1 = self.take_tri_by_mask(
            ii_anch_1, ii_pos_1, ii_neg_1, mask=labels[ii_anch_1] != labels[ii_neg_1]
        )
        n_batch_tri = len(ii_anch_1)

        # part2: anchor + positive come from bank, negative comes from batch
        n_tri_positives_from_bank = int(n_batch_tri * (self.tri_expand_k - 1) / 2)
        ii_anch_2, ii_pos_2 = self.get_pos_pairs(lbl2idx_bank, n_tri_positives_from_bank).T
        ii_neg_2 = randint(0, self.bs, size=(len(ii_anch_2),))

        ii_anch_2, ii_pos_2, ii_neg_2 = self.take_tri_by_mask(
            ii_anch_2, ii_pos_2, ii_neg_2, mask=self.bank_labels[ii_anch_2] != labels[ii_neg_2]
        )

        # part3: anchor + positive come from batch, negative comes from bank
        # we try to make size of this part equals to part2
        n_tri_negatives_from_bank = n_tri_positives_from_bank
        ii_anch_3, ii_pos_3 = self.get_pos_pairs(lbl2idx_batch).T
        ii_neg_3 = randint(0, self.bank_size, size=(n_tri_negatives_from_bank,))
        ii_3 = randint(0, len(ii_anch_3), size=(n_tri_negatives_from_bank,))

        ii_anch_3 = ii_anch_3[ii_3]
        ii_pos_3 = ii_pos_3[ii_3]
        ii_anch_3, ii_pos_3, ii_neg_3 = self.take_tri_by_mask(
            ii_anch_3, ii_pos_3, ii_neg_3, mask=labels[ii_anch_3] != self.bank_labels[ii_neg_3]
        )

        features_anch = cat([features[ii_anch_3], self.bank_features[ii_anch_2], features[ii_anch_1]])
        features_pos = cat([features[ii_pos_3], self.bank_features[ii_pos_2], features[ii_pos_1]])
        features_neg = cat([self.bank_features[ii_neg_3], features[ii_neg_2], features[ii_neg_1]])
        assert len(features_anch) == len(features_pos) == len(features_neg)

        self.update_bank(features=features, labels=labels)

        is_original_tri = zeros(len(features_anch), dtype=bool).cpu()
        is_original_tri[-len(ii_anch_1) :] = True

        # print(len(ii_anch_1), len(ii_anch_2), len(ii_anch_3))

        return features_anch, features_pos, features_neg, is_original_tri

    @staticmethod
    def take_tri_by_mask(ii_a: Tensor, ii_p: Tensor, ii_n: Tensor, mask: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        ii_a = ii_a[mask]
        ii_p = ii_p[mask]
        ii_n = ii_n[mask]
        return ii_a, ii_p, ii_n


__all__ = ["TripletMinerWithMemory"]



import torch
from torch import Tensor

from oml.functional.losses import surrogate_precision
from oml.utils.misc_torch import pairwise_dist


class SurrogatePrecision(torch.nn.Module):
    """
    This loss is a differentiable approximation of Precision@k metric.

    The loss is described in the following paper under a bit different name:
    `Recall@k Surrogate Loss with Large Batches and Similarity Mixup`_.

    The idea is that we express the formula for Precision@k using two step functions (aka Heaviside functions).
    Then we approximate them using two sigmoid functions with temperatures.
    The smaller temperature the close sigmoid to the step function, but the gradients are sparser,
    and vice versa. In the original paper `t1 = 1.0` and `t2 = 0.01` have been used.

    .. _Recall@k Surrogate Loss with Large Batches and Similarity Mixup:
        https://arxiv.org/pdf/2108.11179v2.pdf

    """

    criterion_name = "surrogate_precision"

    def __init__(self, k: int, temperature1: float = 1.0, temperature2: float = 0.01, reduction: str = "mean"):
        """

        Args:
            k: Parameter of Precision@k.
            temperature1: Scaling factor for the 1st sigmoid, see docs above.
            temperature2: Scaling factor for the 2nd sigmoid, see docs above.
            reduction: ``mean``, ``sum`` or ``none``

        """
        super(SurrogatePrecision, self).__init__()
        assert k > 0

        """
        Note, since we consider all the batch samples as queries and galleries simultaneously,
        for each element we have its copy on the 1st position with the corresponding zero distance.
        Thus, to consider it we increase parameter k by 1.
        """

        self.k = k + 1
        self.temperature1 = temperature1
        self.temperature2 = temperature2
        self.reduction = reduction

    def forward(self, features: torch.Tensor, labels: Tensor) -> Tensor:
        """

        Args:
            features: Features with the shape of ``[batch_size, feature_size]``
            labels: Labels with the size of ``batch_size``

        Returns:
            Loss value

        """
        assert len(features) == len(labels)

        distances = pairwise_dist(x1=features, x2=features)
        mask_gt = labels[..., None] == labels[None, ...]

        loss = 1 - surrogate_precision(
            distances=distances,
            mask_gt=mask_gt,
            t1=self.temperature1,
            t2=self.temperature2,
            k=self.k,
            reduction=self.reduction,
        )

        return loss


__all__ = ["SurrogatePrecision"]



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Union

import numpy as np
from torch import LongTensor
from torch.utils.data import Dataset

from oml.const import INPUT_TENSORS_KEY_1, INPUT_TENSORS_KEY_2, LABELS_KEY, TColor


class IIndexedDataset(Dataset, ABC):
    index_key: str

    def __getitem__(self, item: int) -> Dict[str, Any]:
        """

        Args:
            item: Idx of the sample

        Returns:
            Dictionary having the following key:
            ``self.index_key: int = item``

        """
        raise NotImplementedError()

    def __len__(self) -> int:
        raise NotImplementedError()


class IBaseDataset(IIndexedDataset, ABC):
    input_tensors_key: str
    extra_data: Dict[str, Any]  # container for storing extra records having the same size as the dataset

    def __getitem__(self, item: int) -> Dict[str, Any]:
        """

        Args:
            item: Idx of the sample

        Returns:
            Dictionary including the following keys:
            ``self.input_tensors_key``
            ``self.index_key: int = item``

        """
        raise NotImplementedError()


class ILabeledDataset(IBaseDataset, ABC):
    """
    This is an interface for the datasets which provide labels of containing items.

    """

    labels_key: str = LABELS_KEY

    def __getitem__(self, item: int) -> Dict[str, Any]:
        """

        Args:
            item: Idx of the sample

        Returns:
             Dictionary including the following keys:

            ``self.input_tensors_key``
            ``self.index_key: int = item``
            ``self.labels_key``

        """
        raise NotImplementedError()

    @abstractmethod
    def get_labels(self) -> np.ndarray:
        raise NotImplementedError()

    def get_label2category(self) -> Optional[Dict[int, Union[str, int]]]:
        """
        Returns:
            Mapping from label to category if known.

        """
        raise NotImplementedError()


class IQueryGalleryDataset(IBaseDataset, ABC):
    """
    This is an interface for the datasets which hold the information on how to split
    the data into the query and gallery. The query and gallery ids may overlap.
    It doesn't need the ground truth labels, so it can be used for prediction on not annotated data.

    """

    @abstractmethod
    def get_query_ids(self) -> LongTensor:
        raise NotImplementedError()

    @abstractmethod
    def get_gallery_ids(self) -> LongTensor:
        raise NotImplementedError()


class IQueryGalleryLabeledDataset(IQueryGalleryDataset, ILabeledDataset, ABC):
    """
    This interface is similar to `IQueryGalleryDataset`, but there are ground truth labels.
    """


class IPairDataset(IIndexedDataset):
    """
    This is an interface for the datasets which return pair of something.

    """

    input_tensors_key_1: str = INPUT_TENSORS_KEY_1
    input_tensors_key_2: str = INPUT_TENSORS_KEY_2

    @abstractmethod
    def __getitem__(self, item: int) -> Dict[str, Any]:
        """
        Args:
            item: Idx of the sample

        Returns:
             Dictionary with the following keys:

            ``self.input_tensors_key_1``
            ``self.input_tensors_key_2``
            ``self.index_key``

        """
        raise NotImplementedError()


class IVisualizableDataset(Dataset, ABC):
    """
    Base class for the datasets which know how to visualise their items.
    """

    @abstractmethod
    def visualize(self, item: int, color: TColor) -> np.ndarray:
        raise NotImplementedError()


__all__ = [
    "IIndexedDataset",
    "IBaseDataset",
    "ILabeledDataset",
    "IQueryGalleryLabeledDataset",
    "IQueryGalleryDataset",
    "IPairDataset",
    "IVisualizableDataset",
]



import inspect

from oml.registry.losses import LOSSES_REGISTRY
from oml.registry.miners import MINERS_REGISTRY
from oml.registry.models import EXTRACTORS_REGISTRY, PAIRWISE_MODELS_REGISTRY
from oml.registry.optimizers import OPTIMIZERS_REGISTRY
from oml.registry.samplers import SAMPLERS_REGISTRY
from oml.registry.schedulers import SCHEDULERS_REGISTRY
from oml.registry.transforms import TRANSFORMS_REGISTRY, get_transforms_for_pretrained


def show_registry() -> None:
    for name, registry in [
        ("Losses", LOSSES_REGISTRY),
        ("Miners", MINERS_REGISTRY),
        ("Models", EXTRACTORS_REGISTRY),
        ("Optimizers", OPTIMIZERS_REGISTRY),
        ("Samplers", SAMPLERS_REGISTRY),
        ("Schedulers", SCHEDULERS_REGISTRY),
        ("Augmentations", TRANSFORMS_REGISTRY),
    ]:
        print(f"{name}: ")

        if name == "Augmentations":
            for k, constructor in registry.items():  # type: ignore
                print(f"{k}:", constructor.__class__.__name__ + str(inspect.signature(constructor)))

        else:
            for k, constructor in registry.items():  # type: ignore
                print(f"{k}:", constructor.__name__ + str(inspect.signature(constructor)))

        print()



from typing import Any, Dict

from oml.interfaces.miners import ITripletsMiner
from oml.miners.cross_batch import TripletMinerWithMemory
from oml.miners.inbatch_all_tri import AllTripletsMiner
from oml.miners.inbatch_hard_cluster import HardClusterMiner
from oml.miners.inbatch_hard_tri import HardTripletsMiner
from oml.miners.inbatch_nhard_tri import NHardTripletsMiner
from oml.miners.miner_with_bank import MinerWithBank
from oml.utils.misc import TCfg, dictconfig_to_dict

MINERS_REGISTRY = {
    "all_triplets": AllTripletsMiner,
    "hard_cluster": HardClusterMiner,
    "hard_triplets": HardTripletsMiner,
    "triplets_with_memory": TripletMinerWithMemory,
    "miner_with_bank": MinerWithBank,
    "n_hard_triplets": NHardTripletsMiner,
}


def get_miner(name: str, **kwargs: Dict[str, Any]) -> ITripletsMiner:
    if "miner" in kwargs:
        miner = get_miner_by_cfg(kwargs["miner"].copy())
        del kwargs["miner"]
        return MINERS_REGISTRY[name](miner=miner, **kwargs)
    else:
        return MINERS_REGISTRY[name](**kwargs)


def get_miner_by_cfg(cfg: TCfg) -> ITripletsMiner:
    cfg = dictconfig_to_dict(cfg)
    return get_miner(name=cfg["name"], **cfg["args"])


__all__ = ["MINERS_REGISTRY", "get_miner", "get_miner_by_cfg"]



from abc import ABC, abstractmethod
from collections import Counter
from typing import List, Tuple, Union

from torch import Tensor

TTriplets = Tuple[Tensor, Tensor, Tensor]
TTripletsIds = Tuple[List[int], List[int], List[int]]
TLabels = Union[List[int], Tensor]


def labels2list(labels: TLabels) -> List[int]:
    if isinstance(labels, Tensor):
        labels = labels.squeeze()
        labels_list = labels.tolist()
    elif isinstance(labels, list):
        labels_list = labels.copy()
    else:
        raise TypeError(f"Unexpected type of labels: {type(labels)}).")

    return labels_list


class ITripletsMiner(ABC):
    """
    An abstraction of triplet miner.

    """

    @abstractmethod
    def sample(self, features: Tensor, labels: TLabels) -> TTriplets:
        """
        This method includes the logic of mining/sampling triplets.

        Args:
            features: Features with the shape of ``[batch_size, feature_size]``
            labels: Labels with the size of ``batch_size``

        Returns:
            Batch of triplets

        """
        raise NotImplementedError()


class ITripletsMinerInBatch(ITripletsMiner):
    """
    We expect that the child instances of this class
    will be used for mining triplets inside the batches.
    The batches must contain at least 2 samples for
    each class and at least 2 different labels,
    such behaviour can be guarantee via using samplers from
    our registry.

    But you are not limited to using it in any other way.

    """

    @staticmethod
    def _check_input_labels(labels: List[int]) -> None:
        """
        Args:
            labels: Labels of the samples in the batch

        """
        labels_counter = Counter(labels)
        assert all(n > 1 for n in labels_counter.values())
        assert len(labels_counter) > 1

    @abstractmethod
    def _sample(self, features: Tensor, labels: List[int]) -> TTripletsIds:
        """
        This method includes the logic of mining triplets
        inside the batch. It can be based on information about
        the distance between the features, or the
        choice can be performed randomly.

        Args:
            features: Features with the shape of ``[batch_size, feature_size]``
            labels: Labels with the size of ``batch_size``

        Returns:
            Indices of the batch samples to form the triplets

        """
        raise NotImplementedError()

    def sample(self, features: Tensor, labels: TLabels) -> TTriplets:
        """
        Args:
            features: Features with the shape of ``[batch_size, feature_size]``
            labels: Labels with the size of ``batch_size``

        Returns:
             Batch of triplets

        """
        # Convert labels to list
        labels = labels2list(labels)
        self._check_input_labels(labels=labels)

        ids_anchor, ids_pos, ids_neg = self._sample(features, labels=labels)

        return features[ids_anchor], features[ids_pos], features[ids_neg]


__all__ = ["TTriplets", "TTripletsIds", "TLabels", "labels2list", "ITripletsMiner", "ITripletsMinerInBatch"]



from oml.retrieval.postprocessors.algo import AdaptiveThresholding, ConstantThresholding
from oml.retrieval.postprocessors.pairwise import PairwiseReranker
from oml.retrieval.retrieval_results import RetrievalResults



import warnings
from typing import Any

from torch import FloatTensor

from oml.interfaces.models import IExtractor

THFModel = Any
TBatchEncoding = Any


def check_model_type(model):  # type: ignore
    try:
        import transformers as t

        if not isinstance(model, (t.RobertaModel, t.BertModel, t.DistilBertModel)):
            warnings.warn(
                f"Unexpected model type: {type(model)}. Make sure your model is "
                f"compatible with {HFWrapper.__name__}."
            )
    except ImportError:
        pass


class HFWrapper(IExtractor):
    """
    This is a wrapper for models from HuggingFace (transformers) library.
    Please, double check `forward` method to learn how we extrac features.
    """

    def __init__(self, model: THFModel, feat_dim: int):
        super().__init__()
        check_model_type(model)

        self.model = model
        self._feat_dim = feat_dim

    def forward(self, x: TBatchEncoding) -> FloatTensor:  # type: ignore
        hf_output = self.model(x["input_ids"], attention_mask=x["attention_mask"])
        embedding = hf_output["last_hidden_state"][:, 0, :]
        return embedding

    @property
    def feat_dim(self) -> int:
        return self._feat_dim


__all_ = ["HuggingFaceWrapper"]



from typing import List, Tuple

import torch
from torch import Tensor

from oml.interfaces.miners import ITripletsMinerInBatch
from oml.miners.inbatch_all_tri import AllTripletsMiner
from oml.miners.inbatch_hard_tri import HardTripletsMiner


class PairsMiner:
    # todo: add this class to registry, probably add BCEWithPairsMiner as well

    _miner: ITripletsMinerInBatch

    def __init__(self, hard_mining: bool = True):
        super().__init__()
        self._miner = HardTripletsMiner() if hard_mining else AllTripletsMiner()

    def sample(self, features: Tensor, labels: List[int]) -> Tuple[Tensor, Tensor, Tensor]:
        ii_a, ii_p, ii_n = self._miner._sample(features, labels=labels)

        # todo: refactor the current approach when we create pairs disassembling triplets
        ii_a_1, ii_p = zip(*list(set(list(map(lambda x: tuple(sorted([x[0], x[1]])), zip(ii_a, ii_p))))))
        ii_a_2, ii_n = zip(*list(set(list(map(lambda x: tuple(sorted([x[0], x[1]])), zip(ii_a, ii_n))))))

        is_negative = torch.ones(len(ii_a_1) + len(ii_a_2)).bool()
        is_negative[: len(ii_a_1)] = 0

        return torch.tensor([*ii_a_1, *ii_a_2]).long(), torch.tensor([*ii_p, *ii_n]).long(), is_negative


__all__ = ["PairsMiner"]



from oml.datasets.images import (
    ImageBaseDataset,
    ImageLabeledDataset,
    ImageQueryGalleryDataset,
    ImageQueryGalleryLabeledDataset,
)
from oml.datasets.pairs import PairDataset
from oml.datasets.texts import (
    TextBaseDataset,
    TextLabeledDataset,
    TextQueryGalleryDataset,
    TextQueryGalleryLabeledDataset,
)



import io
import re
from pathlib import Path
from typing import List

from setuptools import find_packages, setup


def load_requirements(filename: str) -> List[str]:
    with open(filename, "r") as f:
        reqs = f.read().splitlines()
    return reqs


def load_version() -> str:
    version_file = Path(__file__).parent / "oml" / "__init__.py"
    with io.open(version_file, encoding="utf-8") as f:
        return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)


NLP_REQUIRE = load_requirements("ci/requirements_nlp.txt")


setup(
    # technical things
    version=load_version(),
    packages=find_packages(exclude=["ci", "docs", "pipelines", "tests*"]),
    python_requires=">=3.8,<4.0",
    install_requires=load_requirements("ci/requirements.txt"),
    extras_require={
        "nlp": NLP_REQUIRE,
        "all": NLP_REQUIRE,  # later will be cv and audio
    },
    include_package_data=True,
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    # general information
    name="open-metric-learning",
    description="OML is a PyTorch-based framework to train and validate the models producing high-quality embeddings.",
    keywords=[
        "data-science",
        "computer-vision",
        "deep-learning",
        "pytorch",
        "metric-learning",
        "representation-learning",
        "pytorch-lightning",
    ],
    author="Shabanov Aleksei",
    author_email="shabanoff.aleksei@gmail.com",
    url="https://github.com/OML-Team/open-metric-learning",
    classifiers=[
        "Environment :: Console",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    project_urls={
        "Homepage": "https://github.com/OML-Team/open-metric-learning",
        "Bug Tracker": "https://github.com/OML-Team/open-metric-learning/issues",
    },
    license="Apache License 2.0",
)



import itertools
import warnings
from typing import Any, Dict, List, Optional, Union

import numpy as np
import torch
from torch.distributed import all_gather_object, get_rank, get_world_size


class WarningDDP(UserWarning):
    def __init__(self, function_name: str):
        self.message = (
            f"You use '{function_name}' function, but the default process group has not been initialized. "
            "In this case for compatibility function doesn't modify your data. "
            "Please make sure to call init_process_group or ignore this warning if you want to use this "
            "function without DDP"
        )

    def __str__(self) -> str:
        return repr(self.message)


def merge_list_of_dicts(list_of_dicts: List[Dict[str, Any]], device: torch.device = "cpu") -> Dict[str, Any]:
    available_types = (list, tuple, torch.Tensor, np.ndarray)

    for dict_sample in list_of_dicts:
        assert all(
            isinstance(v, available_types) for v in dict_sample.values()
        ), f"Only '{available_types}' are available for merging"
        assert set((k, type(v)) for k, v in list_of_dicts[0].items()) == set(
            (k, type(v)) for k, v in dict_sample.items()
        ), "Dictionaries in list have to have same keys with same types of values"

    output = {}

    for key, _value_for_type_check in list_of_dicts[0].items():
        if isinstance(_value_for_type_check, (list, tuple)):
            output[key] = type(_value_for_type_check)(itertools.chain(*[g[key] for g in list_of_dicts]))
        elif isinstance(_value_for_type_check, torch.Tensor):
            output[key] = torch.cat([g[key].to(device) for g in list_of_dicts], dim=0)
        elif isinstance(_value_for_type_check, np.ndarray):
            output[key] = np.concatenate(tuple(g[key] for g in list_of_dicts), axis=0)
        else:
            raise TypeError(
                f"Only '{available_types}' are available for merging. "
                f"Got type '{type(list_of_dicts[0][key])}' for key '{key}'"
            )

    return output


def sync_dicts_ddp(
    outputs_from_device: Dict[str, Any], world_size: int, device: Union[torch.device, str] = "cpu"
) -> Dict[str, Any]:
    """
    The function allows you to combine and merge data stored in dictionaries from all devices.
    You can place this function in your code and all the devices upon reaching this function will wait for
    each other to synchronize and merge dictionaries.

    Note:
       The function under the hood pickles all object, converts bytes to tensor, then unpickles them after syncing.
       With `NCCL DDP` backend (the default one) intermediate tensors are stored on CUDA.

    """
    if world_size >= 1 and is_ddp():
        gathered: List[Optional[Dict[str, Any]]] = [None for _ in range(world_size)]
        all_gather_object(gathered, outputs_from_device, group=torch.distributed.group.WORLD)
        return merge_list_of_dicts(gathered, device)
    else:
        if world_size > 1:
            warnings.warn(sync_dicts_ddp.__name__, WarningDDP)
        return outputs_from_device


def is_ddp() -> bool:
    try:
        get_world_size()
        return True
    except RuntimeError:
        return False


def get_world_size_safe() -> int:
    try:
        return get_world_size()
    except RuntimeError:
        return 0


def is_main_process() -> bool:
    if is_ddp():
        return get_rank() == 0
    else:
        return True


__all__ = ["is_ddp", "sync_dicts_ddp", "merge_list_of_dicts", "is_main_process"]



from typing import Any, Dict, List, Union

import albumentations as albu
import cv2
import numpy as np
from albumentations.pytorch import ToTensorV2

from oml.const import CROP_KEY, MEAN, MEAN_CLIP, PAD_COLOR, STD, STD_CLIP, TNormParam

TTransformsList = List[Union[albu.ImageOnlyTransform, albu.DualTransform]]


def get_spatials() -> TTransformsList:
    spatial_augs = [
        albu.Perspective(scale=(0.06, 0.07), pad_mode=cv2.BORDER_CONSTANT, pad_val=PAD_COLOR),
        albu.Affine(
            scale=None,
            rotate=(-0.1, +0.1),
            translate_percent=(-0.05, 0.05),
            shear=(-7, 7),
            cval=PAD_COLOR,
        ),
        albu.GridDistortion(
            distort_limit=0.1,
            border_mode=0,
            value=PAD_COLOR,
        ),
    ]
    return spatial_augs


def get_blurs() -> TTransformsList:
    blur_augs = [
        albu.MotionBlur(),
        albu.MedianBlur(),
        albu.Blur(),
        albu.GaussianBlur(sigma_limit=(0.7, 2.0)),
    ]
    return blur_augs


def get_colors_level() -> TTransformsList:
    color_augs = [
        albu.ColorJitter(brightness=0.4, contrast=0.4, saturation=0.4, hue=0.1, p=0.5),
        albu.HueSaturationValue(p=0.1),
        albu.CLAHE(p=0.1),
        albu.Sharpen(p=0.1),
        albu.Emboss(p=0.1),
        albu.RandomBrightnessContrast(p=0.1),
    ]
    return color_augs


def get_noises() -> TTransformsList:
    noise_augs = [
        albu.CoarseDropout(max_holes=3, max_height=20, max_width=20, fill_value=PAD_COLOR, p=0.3),
        albu.GaussNoise(p=0.7),
    ]
    return noise_augs


def get_noise_channels() -> TTransformsList:
    channels_noise_augs = [
        albu.ChannelDropout(p=0.1),
        albu.ToGray(p=0.8),
        albu.ChannelShuffle(p=0.1),
    ]
    return channels_noise_augs


class Crop:
    def __init__(self, crop_key: str = CROP_KEY):
        self.crop_key = crop_key

    def __call__(self, image: np.ndarray, **kwargs: Dict[str, Any]) -> Dict[str, np.ndarray]:
        x1, y1, x2, y2 = kwargs[self.crop_key]
        return {"image": image[y1:y2, x1:x2, :]}  # type: ignore

    def _to_dict(self) -> Dict[str, Any]:
        return {"crop_key": self.crop_key}


class RandomSizedBBoxSafeCropPatched:
    """
    The only difference is from using the original albu.RandomSizedBBoxSafeCrop is that
    we do not pass any box information after cropping which allows us to avoid limitations
    in using further augmentations.
    """

    def __init__(self, size: int, erosion_rate: float, crop_key: str = CROP_KEY):
        self.crop_key = crop_key
        self.transform = albu.Compose(
            [albu.RandomSizedBBoxSafeCrop(width=size, height=size, erosion_rate=erosion_rate)],
            bbox_params=albu.BboxParams(format="pascal_voc", label_fields=["label_field"]),
        )

    def __call__(self, image: np.ndarray, **kwargs: Dict[str, Any]) -> Dict[str, np.ndarray]:
        x1, y1, x2, y2 = kwargs[self.crop_key]
        output = self.transform(image=image, bboxes=[[x1, y1, x2, y2]], label_field=[1])
        return {"image": output["image"]}

    def _to_dict(self) -> Dict[str, Any]:
        return {**self.transform._to_dict(), **{"crop_key": self.crop_key}}


def get_augs_albu(im_size: int, mean: TNormParam = MEAN, std: TNormParam = STD) -> albu.Compose:
    augs = albu.Compose(
        [
            albu.LongestMaxSize(max_size=im_size),
            albu.PadIfNeeded(min_height=im_size, min_width=im_size, border_mode=cv2.BORDER_CONSTANT, value=PAD_COLOR),
            albu.HorizontalFlip(p=0.5),
            albu.OneOf(get_spatials(), p=0.5),
            albu.OneOf(get_blurs(), p=0.5),
            albu.OneOf(get_colors_level(), p=0.8),
            albu.OneOf(get_noise_channels(), p=0.2),
            albu.OneOf(get_noises(), p=0.25),
            albu.Normalize(mean=mean, std=std),
            ToTensorV2(),
        ],
    )
    return augs


def get_normalisation_albu(mean: TNormParam = MEAN, std: TNormParam = STD) -> albu.Compose:
    return albu.Compose([albu.Normalize(mean=mean, std=std), ToTensorV2()])


def get_normalisation_resize_albu(im_size: int, mean: TNormParam = MEAN, std: TNormParam = STD) -> albu.Compose:
    return albu.Compose(
        [
            albu.LongestMaxSize(max_size=im_size),
            albu.PadIfNeeded(min_height=im_size, min_width=im_size, border_mode=cv2.BORDER_CONSTANT, value=PAD_COLOR),
            albu.Normalize(mean=mean, std=std),
            ToTensorV2(),
        ]
    )


def get_normalisation_resize_albu_clip(im_size: int) -> albu.Compose:
    return get_normalisation_resize_albu(im_size=im_size, mean=MEAN_CLIP, std=STD_CLIP)


__all__ = [
    "get_augs_albu",
    "get_normalisation_albu",
    "get_normalisation_resize_albu",
    "get_normalisation_resize_albu_clip",
    "get_spatials",
    "get_blurs",
    "get_colors_level",
    "get_noise_channels",
    "get_noises",
    "RandomSizedBBoxSafeCropPatched",
]



"""
Mostly copy-paste from ruclip library.
https://github.com/openai/CLIP/blob/main/clip/model.py
"""

from collections import OrderedDict

import torch
from torch import nn


class LayerNorm(nn.LayerNorm):
    """Subclass torch's LayerNorm to handle fp16."""

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        orig_type = x.dtype
        ret = super().forward(x.type(torch.float32))
        return ret.type(orig_type)


class QuickGELU(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * torch.sigmoid(1.702 * x)


class ResidualAttentionBlock(nn.Module):
    def __init__(self, d_model: int, n_head: int, attn_mask: torch.Tensor = None):
        super().__init__()

        self.attn = nn.MultiheadAttention(d_model, n_head)
        self.ln_1 = LayerNorm(d_model)
        self.mlp = nn.Sequential(
            OrderedDict(
                [
                    ("c_fc", nn.Linear(d_model, d_model * 4)),
                    ("gelu", QuickGELU()),
                    ("c_proj", nn.Linear(d_model * 4, d_model)),
                ]
            )
        )
        self.ln_2 = LayerNorm(d_model)
        self.attn_mask = attn_mask

    def attention(self, x: torch.Tensor) -> torch.Tensor:
        self.attn_mask = self.attn_mask.to(dtype=x.dtype, device=x.device) if self.attn_mask is not None else None
        return self.attn(x, x, x, need_weights=False, attn_mask=self.attn_mask)[0]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.attention(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))
        return x


class Transformer(nn.Module):
    def __init__(self, width: int, layers: int, heads: int, attn_mask: torch.Tensor = None):
        super().__init__()
        self.width = width
        self.layers = layers
        self.resblocks = nn.Sequential(*[ResidualAttentionBlock(width, heads, attn_mask) for _ in range(layers)])

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.resblocks(x)


class VisionTransformer(nn.Module):
    def __init__(self, input_resolution: int, patch_size: int, width: int, layers: int, heads: int, output_dim: int):
        super().__init__()
        self.input_resolution = input_resolution
        self.output_dim = output_dim
        self.conv1 = nn.Conv2d(in_channels=3, out_channels=width, kernel_size=patch_size, stride=patch_size, bias=False)

        scale = width**-0.5
        self.class_embedding = nn.Parameter(scale * torch.randn(width))
        self.positional_embedding = nn.Parameter(scale * torch.randn((input_resolution // patch_size) ** 2 + 1, width))
        self.ln_pre = LayerNorm(width)

        self.transformer = Transformer(width, layers, heads)

        self.ln_post = LayerNorm(width)
        self.proj = nn.Parameter(scale * torch.randn(width, output_dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)  # shape = [*, width, grid, grid]
        x = x.reshape(x.shape[0], x.shape[1], -1)  # shape = [*, width, grid ** 2]
        x = x.permute(0, 2, 1)  # shape = [*, grid ** 2, width]
        x = torch.cat(
            [
                self.class_embedding.to(x.dtype)
                + torch.zeros(x.shape[0], 1, x.shape[-1], dtype=x.dtype, device=x.device),
                x,
            ],
            dim=1,
        )  # shape = [*, grid ** 2 + 1, width]
        x = x + self.positional_embedding.to(x.dtype)
        x = self.ln_pre(x)

        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)
        x = x.permute(1, 0, 2)  # LND -> NLD

        x = self.ln_post(x[:, 0, :])

        if self.proj is not None:
            x = x @ self.proj

        return x


__all__ = ["QuickGELU", "VisionTransformer"]



import multiprocessing as mp
from functools import partial
from io import BytesIO
from pathlib import Path
from typing import Callable, List, Optional, Union

import matplotlib.pyplot as plt
from tqdm import tqdm

try:
    from albumentations.augmentations.functional import pad
except (AttributeError, ModuleNotFoundError, ImportError):
    from albumentations.augmentations.geometric.functional import pad

import cv2
import numpy as np
import PIL
import torch
from PIL import Image
from PIL.Image import Image as TPILImage

from oml.const import PAD_COLOR, TBBox, TColor

TImage = Union[PIL.Image.Image, np.ndarray]
TImReader = Callable[[Union[Path, str, bytes]], TImage]


def tensor_to_numpy_image(img: torch.Tensor) -> np.ndarray:
    img = img.detach().cpu()
    img = img.permute(1, 2, 0).numpy().copy()

    img -= img.min()
    img /= img.max()
    img *= 255

    img = img.astype(np.uint8)

    return img


def imread_cv2(im_src: Union[Path, str, bytes]) -> np.ndarray:
    if isinstance(im_src, (Path, str)):
        image = cv2.imread(str(im_src), cv2.IMREAD_UNCHANGED)
    elif isinstance(im_src, bytes):
        image_raw = np.frombuffer(im_src, np.uint8)
        image = cv2.imdecode(image_raw, cv2.IMREAD_UNCHANGED)
    else:
        raise TypeError("Unsupported type")

    return cv2.cvtColor(image, cv2.COLOR_BGRA2RGB)


def imread_pillow(im_src: Union[Path, str, bytes]) -> TPILImage:
    if isinstance(im_src, (Path, str)):
        image = Image.open(im_src)
    elif isinstance(im_src, bytes):
        image = Image.open(BytesIO(im_src))
    else:
        raise TypeError("Unsupported type")
    return image.convert("RGB")


def draw_bbox(im: np.ndarray, bbox: Optional[TBBox], color: TColor) -> np.ndarray:
    """
    Draws a single bounding box on the image.
    If the elements of the bbox are NaNs, we will draw bbox around the whole image.

    Args:
        im: Image
        bbox: Optional single bounding box in the format of ``[x1, y1, x2, y2]``.
        color: Tuple of 3 ints
    """
    im_ret = im.copy()

    if bbox is None:
        x1, y1, x2, y2 = 0, 0, im_ret.shape[1], im_ret.shape[0]
    else:
        x1, y1, x2, y2 = bbox

    im_avg_sz = (im_ret.shape[0] + im_ret.shape[1]) / 2
    thickness = max(3, int(0.05 * im_avg_sz))
    im_ret = cv2.rectangle(im_ret, (x1, y1), (x2, y2), thickness=thickness, color=color)

    return im_ret


def square_pad(img: np.ndarray) -> np.ndarray:
    return pad(img, min_height=max(img.shape), min_width=max(img.shape), border_mode=0, value=PAD_COLOR)


def try_to_open_image(im_path: Path, f_imread: TImReader) -> Optional[str]:
    try:
        _ = f_imread(im_path)
        return None
    except Exception:
        return str(im_path)


def find_broken_images(images_list: List[Path], f_imread: TImReader, num_processes: int = 10) -> List[str]:
    try_to_open_image_ = partial(try_to_open_image, f_imread=f_imread)

    with mp.Pool(processes=num_processes) as p:
        results = list(tqdm(p.imap(try_to_open_image_, images_list), total=len(images_list)))

    results = list(filter(lambda x: x is not None, results))
    return results


def figure_to_nparray(fig: plt.Figure) -> np.ndarray:
    """
    Converts a matplotlib figure to a numpy array with RGB channels and no alpha channel.
    """
    fig.canvas.draw()
    data = np.array(fig.canvas.renderer.buffer_rgba())[..., :3]
    plt.close(fig)
    return data


__all__ = [
    "TImage",
    "TImReader",
    "tensor_to_numpy_image",
    "imread_cv2",
    "imread_pillow",
    "draw_bbox",
    "square_pad",
    "figure_to_nparray",
]



# type: ignore
# flake8: noqa

# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch
from torchvision.models.resnet import resnet50

import oml.models.vit_dino.external.vision_transformer as vits

dependencies = ["torch", "torchvision"]


def dino_vits16(pretrained=True, **kwargs):
    """
    ViT-Small/16x16 pre-trained with DINO.
    Achieves 74.5% top-1 accuracy on ImageNet with k-NN classification.
    """
    model = vits.__dict__["vit_small"](patch_size=16, num_classes=0, **kwargs)
    if pretrained:
        state_dict = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/dino/dino_deitsmall16_pretrain/dino_deitsmall16_pretrain.pth",
            map_location="cpu",
        )
        model.load_state_dict(state_dict, strict=True)
    return model


def dino_vits8(pretrained=True, **kwargs):
    """
    ViT-Small/8x8 pre-trained with DINO.
    Achieves 78.3% top-1 accuracy on ImageNet with k-NN classification.
    """
    model = vits.__dict__["vit_small"](patch_size=8, num_classes=0, **kwargs)
    if pretrained:
        state_dict = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/dino/dino_deitsmall8_pretrain/dino_deitsmall8_pretrain.pth",
            map_location="cpu",
        )
        model.load_state_dict(state_dict, strict=True)
    return model


def dino_vitb16(pretrained=True, **kwargs):
    """
    ViT-Base/16x16 pre-trained with DINO.
    Achieves 76.1% top-1 accuracy on ImageNet with k-NN classification.
    """
    model = vits.__dict__["vit_base"](patch_size=16, num_classes=0, **kwargs)
    if pretrained:
        state_dict = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/dino/dino_vitbase16_pretrain/dino_vitbase16_pretrain.pth",
            map_location="cpu",
        )
        model.load_state_dict(state_dict, strict=True)
    return model


def dino_vitb8(pretrained=True, **kwargs):
    """
    ViT-Base/8x8 pre-trained with DINO.
    Achieves 77.4% top-1 accuracy on ImageNet with k-NN classification.
    """
    model = vits.__dict__["vit_base"](patch_size=8, num_classes=0, **kwargs)
    if pretrained:
        state_dict = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/dino/dino_vitbase8_pretrain/dino_vitbase8_pretrain.pth",
            map_location="cpu",
        )
        model.load_state_dict(state_dict, strict=True)
    return model


def dino_resnet50(pretrained=True, **kwargs):
    """
    ResNet-50 pre-trained with DINO.
    Achieves 75.3% top-1 accuracy on ImageNet linear evaluation benchmark (requires to train `fc`).
    """
    model = resnet50(pretrained=False, **kwargs)
    model.fc = torch.nn.Identity()
    if pretrained:
        state_dict = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/dino/dino_resnet50_pretrain/dino_resnet50_pretrain.pth",
            map_location="cpu",
        )
        model.load_state_dict(state_dict, strict=False)
    return model


def dino_xcit_small_12_p16(pretrained=True, **kwargs):
    """
    XCiT-Small-12/16 pre-trained with DINO.
    """
    model = torch.hub.load("facebookresearch/xcit:main", "xcit_small_12_p16", num_classes=0, **kwargs)
    if pretrained:
        state_dict = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/dino/dino_xcit_small_12_p16_pretrain/dino_xcit_small_12_p16_pretrain.pth",
            map_location="cpu",
        )
        model.load_state_dict(state_dict, strict=True)
    return model


def dino_xcit_small_12_p8(pretrained=True, **kwargs):
    """
    XCiT-Small-12/8 pre-trained with DINO.
    """
    model = torch.hub.load("facebookresearch/xcit:main", "xcit_small_12_p8", num_classes=0, **kwargs)
    if pretrained:
        state_dict = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/dino/dino_xcit_small_12_p8_pretrain/dino_xcit_small_12_p8_pretrain.pth",
            map_location="cpu",
        )
        model.load_state_dict(state_dict, strict=True)
    return model


def dino_xcit_medium_24_p16(pretrained=True, **kwargs):
    """
    XCiT-Medium-24/16 pre-trained with DINO.
    """
    model = torch.hub.load("facebookresearch/xcit:main", "xcit_medium_24_p16", num_classes=0, **kwargs)
    if pretrained:
        state_dict = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/dino/dino_xcit_medium_24_p16_pretrain/dino_xcit_medium_24_p16_pretrain.pth",
            map_location="cpu",
        )
        model.load_state_dict(state_dict, strict=True)
    return model


def dino_xcit_medium_24_p8(pretrained=True, **kwargs):
    """
    XCiT-Medium-24/8 pre-trained with DINO.
    """
    model = torch.hub.load("facebookresearch/xcit:main", "xcit_medium_24_p8", num_classes=0, **kwargs)
    if pretrained:
        state_dict = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/dino/dino_xcit_medium_24_p8_pretrain/dino_xcit_medium_24_p8_pretrain.pth",
            map_location="cpu",
        )
        model.load_state_dict(state_dict, strict=True)
    return model



from oml.models.meta.projection import ExtractorWithMLP
from oml.models.meta.siamese import (
    ConcatSiamese,
    LinearTrivialDistanceSiamese,
    TrivialDistanceSiamese,
)
from oml.models.resnet.extractor import ResnetExtractor
from oml.models.texts.huggingface import HFWrapper
from oml.models.vit_clip.extractor import ViTCLIPExtractor
from oml.models.vit_dino.extractor import ViTExtractor
from oml.models.vit_unicom.extractor import ViTUnicomExtractor



from pathlib import Path
from typing import List, Optional, Union

import torch
from torchvision.ops import MLP

from oml.const import STORAGE_CKPTS
from oml.interfaces.models import IExtractor, IFreezable
from oml.models.utils import (
    remove_criterion_in_state_dict,
    remove_prefix_from_state_dict,
)
from oml.models.vit_dino.extractor import ViTExtractor
from oml.utils.io import download_checkpoint


class ExtractorWithMLP(IExtractor, IFreezable):
    """
    Class-wrapper for extractors which an additional MLP.

    """

    pretrained_models = {
        "vits16_224_mlp_384_inshop": {  # type: ignore
            "url": f"{STORAGE_CKPTS}/inshop/vits16_224_mlp_384_inshop.ckpt",
            "hash": "35244966",
            "fname": "vits16_224_mlp_384_inshop.ckpt",
            "init_args": {
                "extractor_creator": lambda: ViTExtractor(None, "vits16", False, use_multi_scale=False),
                "mlp_features": [384],
                "train_backbone": True,
            },
        }
    }

    def __init__(
        self,
        extractor: IExtractor,
        mlp_features: List[int],
        weights: Optional[Union[str, Path]] = None,
        train_backbone: bool = False,
    ):
        """
        Args:
            extractor: Instance of ``IExtractor`` (e.g. ``ViTExtractor``)
            mlp_features: Sizes of projection layers
            weights: Path to weights file or ``None`` for random initialization
            train_backbone: set ``False`` if you want to train only MLP head

        """
        IExtractor.__init__(self)

        self.extractor = extractor
        self.mlp_features = mlp_features
        self.train_backbone = train_backbone

        self.projection = MLP(self.extractor.feat_dim, self.mlp_features)

        if weights:
            if weights in self.pretrained_models:
                pretrained = self.pretrained_models[weights]  # type: ignore
                weights = download_checkpoint(
                    url_or_fid=pretrained["url"],  # type: ignore
                    hash_md5=pretrained["hash"],  # type: ignore
                    fname=pretrained["fname"],  # type: ignore
                )

            loaded = torch.load(weights, map_location="cpu")
            loaded = loaded.get("state_dict", loaded)
            loaded = remove_criterion_in_state_dict(loaded)
            loaded = remove_prefix_from_state_dict(loaded, trial_key="extractor.")
            self.load_state_dict(loaded, strict=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        with torch.set_grad_enabled(self.train_backbone):
            features = self.extractor(x)

        return self.projection(features)

    @property
    def feat_dim(self) -> int:
        return self.mlp_features[-1]

    def freeze(self) -> None:
        self.train_backbone = False

    def unfreeze(self) -> None:
        self.train_backbone = True

    @classmethod
    def from_pretrained(cls, weights: str, **kwargs) -> "IExtractor":  # type: ignore
        # The current class takes another model as a constructor's argument, so, they need to be
        # in the `self.pretrained_models`. The problem is these models will be instantiated even if we simply
        # import something from the current module. To avoid it we added the logic of wrapping/unwrapping
        # constructors into lambda functions.

        ini = cls.pretrained_models[weights]["init_args"]
        ini["extractor"] = ini.pop("extractor_creator")()  # type: ignore

        return super().from_pretrained(weights, **kwargs)


__all__ = ["ExtractorWithMLP"]



import json
from pathlib import Path
from typing import Optional

import requests
from tqdm.auto import tqdm

from oml.const import REQUESTS_TIMEOUT, STORAGE_URL
from oml.utils.io import download_file_from_url


def _fix_path_for_remote_storage(path: str) -> str:
    if path == "/":
        path = ""
    elif path.startswith("/"):
        path = path[1:]

    if path.endswith("/"):
        path = path[:-1]

    return path


def download_file_from_remote_storage(
    remote_path: str, fname: Optional[str] = None, timeout: float = REQUESTS_TIMEOUT
) -> Optional[bytes]:
    remote_path = _fix_path_for_remote_storage(remote_path)
    return download_file_from_url(f"{STORAGE_URL}/download/{remote_path}", fname, timeout=timeout)


def download_folder_from_remote_storage(
    remote_folder: str, local_folder: str, timeout: float = REQUESTS_TIMEOUT
) -> None:
    remote_folder = _fix_path_for_remote_storage(remote_folder)

    remote_ls_response = requests.get(f"{STORAGE_URL}/ls/{remote_folder}", timeout=timeout)
    content = json.loads(remote_ls_response.content)

    if remote_ls_response.status_code == 200:
        remote_files = content["remote_files"]
    else:
        raise Exception(content["detail"])

    local_folder = Path(local_folder)
    local_folder.mkdir(parents=True, exist_ok=True)

    pbar = tqdm(remote_files)
    for remote_file in pbar:
        fname = Path(remote_file).relative_to(remote_folder)
        pbar.set_description_str(f"Downloading '{fname}'")

        local_fname = local_folder / fname
        local_fname.parent.mkdir(exist_ok=True, parents=True)
        download_file_from_remote_storage(remote_file, str(local_fname), timeout=timeout)


__all__ = ["download_folder_from_remote_storage", "download_file_from_remote_storage"]



from typing import Any, Dict
from warnings import warn

from torch import nn

from oml.interfaces.models import IExtractor, IPairwiseModel
from oml.models.meta.projection import ExtractorWithMLP
from oml.models.meta.siamese import (
    ConcatSiamese,
    LinearTrivialDistanceSiamese,
    TrivialDistanceSiamese,
)
from oml.models.resnet.extractor import ResnetExtractor
from oml.models.vit_clip.extractor import ViTCLIPExtractor
from oml.models.vit_dino.extractor import ViTExtractor
from oml.models.vit_unicom.extractor import ViTUnicomExtractor
from oml.utils.misc import TCfg, dictconfig_to_dict

EXTRACTORS_REGISTRY = {
    "resnet": ResnetExtractor,
    "vit": ViTExtractor,
    "vit_clip": ViTCLIPExtractor,
    "vit_unicom": ViTUnicomExtractor,
    "extractor_with_mlp": ExtractorWithMLP,
}

PAIRWISE_MODELS_REGISTRY = {
    "linear_siamese": LinearTrivialDistanceSiamese,
    "concat_siamese": ConcatSiamese,
    "trivial_distance_siamese": TrivialDistanceSiamese,
}


def raise_if_needed(extractor_cfg: TCfg, kwargs: Dict[str, Any], model_name: str) -> None:
    if extractor_cfg.get("weights", False) and kwargs.get("weights", False):
        raise ValueError("You should only provide one weight for extractor_with_mlp.")
    elif extractor_cfg.get("weights", "") is None and kwargs.get("weights", False):
        warn(f"There are weights provided for {model_name}. They can overwrite internal extractor's weights.")


def _get_model(model_name: str, registry: Dict[str, Any], **kwargs: Dict[str, Any]) -> nn.Module:
    if "extractor" in kwargs:  # for nested models
        extractor_cfg = kwargs.pop("extractor")
        inside_extractor = get_extractor_by_cfg(extractor_cfg)
        model = registry[model_name](extractor=inside_extractor, **kwargs)  # type: ignore
        raise_if_needed(extractor_cfg, kwargs, model_name)
    else:
        model = registry[model_name](**kwargs)  # type: ignore
    return model


def get_extractor(model_name: str, **kwargs: Dict[str, Any]) -> IExtractor:
    return _get_model(model_name=model_name, registry=EXTRACTORS_REGISTRY, **kwargs)


def get_extractor_by_cfg(cfg: TCfg) -> IExtractor:
    cfg = dictconfig_to_dict(cfg)
    extractor = get_extractor(model_name=cfg["name"], **cfg["args"])
    return extractor


def get_pairwise_model(model_name: str, **kwargs: Dict[str, Any]) -> IPairwiseModel:
    return _get_model(model_name=model_name, registry=PAIRWISE_MODELS_REGISTRY, **kwargs)


def get_pairwise_model_by_cfg(cfg: TCfg) -> IPairwiseModel:
    cfg = dictconfig_to_dict(cfg)
    pairwise_model = get_pairwise_model(model_name=cfg["name"], **cfg["args"])
    return pairwise_model


__all__ = [
    "EXTRACTORS_REGISTRY",
    "PAIRWISE_MODELS_REGISTRY",
    "get_extractor",
    "get_extractor_by_cfg",
    "get_pairwise_model",
    "get_pairwise_model_by_cfg",
]



from abc import abstractmethod
from typing import Any


class IRetrievalPostprocessor:
    """
    This is a base interface for the classes which somehow postprocess retrieval results.

    """

    @abstractmethod
    def process(self, *args, **kwargs) -> Any:  # type: ignore
        raise NotImplementedError()


__all__ = ["IRetrievalPostprocessor"]
